import os
import re
import tempfile

from pydal import DAL, Field
from pydal._compat import integer_types

from ._adapt import DEFAULT_URI, IS_IMAP, IS_NOSQL, drop
from ._compat import unittest

long = integer_types[-1]

regex_isint = re.compile(r"^[+-]?\d+$")


def range_error_message(error_message, what_to_enter, minimum, maximum):
    "build the error message for the number range validators"
    if error_message is None:
        error_message = "Enter " + what_to_enter
        if minimum is not None and maximum is not None:
            error_message += " between %(min)g and %(max)g"
        elif minimum is not None:
            error_message += " greater than or equal to %(min)g"
        elif maximum is not None:
            error_message += " less than or equal to %(max)g"
    if type(maximum) in integer_types:
        maximum -= 1
    return error_message % dict(min=minimum, max=maximum)


class IS_INT_IN_RANGE(object):
    def __init__(self, minimum=None, maximum=None, error_message=None):
        self.minimum = int(minimum) if minimum is not None else None
        self.maximum = int(maximum) if maximum is not None else None
        self.error_message = range_error_message(
            error_message, "an integer", self.minimum, self.maximum
        )

    def __call__(self, value, record_id=None):
        if regex_isint.match(str(value)):
            v = int(value)
            if (self.minimum is None or v >= self.minimum) and (
                self.maximum is None or v < self.maximum
            ):
                return (v, None)
        return (value, self.error_message)


@unittest.skipIf(IS_IMAP, "TODO: IMAP test")
class TestValidateAndInsert(unittest.TestCase):
    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=["all"])
        db.define_table(
            "val_and_insert",
            Field("aa"),
            Field("bb", "integer", requires=IS_INT_IN_RANGE(1, 5)),
            Field("cc", writable=False),
        )
        rtn = db.val_and_insert.validate_and_insert(aa="test1", bb=2, cc="not-alloed")
        self.assertTrue("cc" in rtn["errors"])
        rtn = db.val_and_insert.validate_and_insert(aa="test1", bb=2)
        if IS_NOSQL:
            self.assertEqual(isinstance(rtn.get("id"), long), True)
        else:
            self.assertEqual(rtn.get("id"), 1)
        # errors should be empty
        self.assertTrue(not rtn.get("errors"))
        # this insert won't pass
        rtn = db.val_and_insert.validate_and_insert(bb="a")
        # the returned id should be None
        self.assertEqual(rtn.get("id"), None)
        # an error message should be in rtn.errors.bb
        self.assertNotEqual(rtn["errors"]["bb"], None)
        # cleanup table
        drop(db.val_and_insert)

    def testApiUpload(self):
        db = DAL(DEFAULT_URI, check_reserved=["all"])
        with tempfile.TemporaryDirectory() as tempdir:
            db.define_table(
                "thing",
                Field("name", required=True),
                Field("image", "upload", uploadfolder=tempdir),
            )
            b64 = b"iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNk+A8AAQUBAScY42YAAAAASUVORK5CYII="
            res = db.thing.validate_and_insert(
                image=dict(filename="image", content=b64)
            )
            self.assertEqual(res, {"id": None, "errors": {"name": "required"}})
            self.assertEqual(os.listdir(tempdir), [])
            res = db.thing.validate_and_insert(
                name="something", image=dict(filename="image", content=b64)
            )
            self.assertEqual(res, {"errors": {}, "id": 1})
            thing = db.thing(res["id"])
            self.assertEqual(thing.name, "something")
            names = os.listdir(tempdir)
            self.assertEqual(len(names), 1)
            self.assertEqual(thing.image, names[0])
            content = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x04\x00\x00\x00\xb5\x1c\x0c\x02\x00\x00\x00\x0bIDATx\xdacd\xf8\x0f\x00\x01\x05\x01\x01'\x18\xe3f\x00\x00\x00\x00IEND\xaeB`\x82"
            with open(os.path.join(tempdir, names[0]), "rb") as stream:
                self.assertEqual(stream.read(), content)


@unittest.skipIf(IS_IMAP, "TODO: IMAP test")
class TestValidateUpdateInsert(unittest.TestCase):
    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=["all"])
        t1 = db.define_table(
            "t1", Field("int_level", "integer", requires=IS_INT_IN_RANGE(1, 5))
        )
        i_response = t1.validate_and_update_or_insert((t1.int_level == 1), int_level=1)
        u_response = t1.validate_and_update_or_insert((t1.int_level == 1), int_level=2)
        e_response = t1.validate_and_update_or_insert((t1.int_level == 1), int_level=6)
        self.assertTrue(i_response["id"] != None)
        self.assertTrue(u_response["id"] != None)
        self.assertTrue(e_response.get("id") == None and e_response.get("errors"))
        self.assertEqual(len(db(t1).select()), 1)
        self.assertEqual(db(t1).count(), 1)
        self.assertEqual(db(t1.int_level == 1).count(), 0)
        self.assertEqual(db(t1.int_level == 6).count(), 0)
        self.assertEqual(db(t1.int_level == 2).count(), 1)
        drop(db.t1)
        return



import os

DEFAULT_URI = os.getenv("DB", "sqlite:memory")
IS_IMAP = "imap" in DEFAULT_URI
IS_GAE = "datastore" in DEFAULT_URI
IS_MONGODB = "mongodb" in DEFAULT_URI
IS_POSTGRESQL = "postgres" in DEFAULT_URI
IS_SQLITE = "sqlite" in DEFAULT_URI
IS_MSSQL = "mssql" in DEFAULT_URI
IS_MYSQL = "mysql" in DEFAULT_URI
IS_TERADATA = "teradata" in DEFAULT_URI
IS_ORACLE = "oracle" in DEFAULT_URI
IS_NOSQL = IS_GAE or IS_MONGODB or IS_IMAP


def drop(table, cascade=None):
    if IS_NOSQL and not IS_MONGODB:
        # GAE drop/cleanup is not implemented
        db = table._db
        db[table]._common_filter = None
        db(table).delete()
        del db[table._tablename]
        del db.tables[db.tables.index(table._tablename)]
        db._remove_references_to(table)
    else:
        if cascade:
            table.drop(cascade)
        else:
            table.drop()


def _quote(db, value):
    return db._adapter.dialect.__class__.quote_template % value



# -*- coding: utf-8 -*-

DRIVERS = {}

try:
    from google.cloud import firestore
    from google.cloud.firestore_v1.base_query import FieldFilter

    DRIVERS["firestore"] = firestore
    psycopg2_adapt = None
    cx_Oracle = None
    pyodbc = None
    couchdb = None
    is_jdbc = False
except ImportError:
    pass

try:
    from pysqlite2 import dbapi2 as sqlite2

    DRIVERS["sqlite2"] = sqlite2
except ImportError:
    pass

try:
    from sqlite3 import dbapi2 as sqlite3

    DRIVERS["sqlite3"] = sqlite3
except ImportError:
    pass

try:
    import pymysql

    DRIVERS["pymysql"] = pymysql
except ImportError:
    pass

try:
    import snowflake.connector as snowflakeconnector

    DRIVERS["snowflakeconnector"] = snowflakeconnector

except ImportError:
    pass

try:
    import MySQLdb

    DRIVERS["MySQLdb"] = MySQLdb
except ImportError:
    pass

try:
    import mysql.connector as mysqlconnector

    DRIVERS["mysqlconnector"] = mysqlconnector
except ImportError:
    pass

try:
    import psycopg2
    from psycopg2.extensions import adapt as psycopg2_adapt

    DRIVERS["psycopg2"] = psycopg2
except ImportError:
    psycopg2_adapt = None

try:
    import cx_Oracle

    DRIVERS["cx_Oracle"] = cx_Oracle
except ImportError:
    cx_Oracle = None

try:
    import pyodbc

    DRIVERS["pyodbc"] = pyodbc
    # DRIVERS.append('DB2(pyodbc)')
    # DRIVERS.append('Teradata(pyodbc)')
    # DRIVERS.append('Ingres(pyodbc)')
except ImportError:
    try:
        import pypyodbc as pyodbc

        DRIVERS["pyodbc"] = pyodbc
    except ImportError:
        pyodbc = None

try:
    import ibm_db_dbi

    DRIVERS["ibm_db_dbi"] = ibm_db_dbi
except ImportError:
    pass
try:
    import Sybase

    DRIVERS["Sybase"] = Sybase
except ImportError:
    pass

try:
    import kinterbasdb

    DRIVERS["kinterbasdb"] = kinterbasdb
    # DRIVERS.append('Firebird(kinterbasdb)')
except ImportError:
    pass

try:
    import fdb

    DRIVERS["fdb"] = fdb
except ImportError:
    pass

try:
    import firebirdsql

    DRIVERS["firebirdsql"] = firebirdsql
except ImportError:
    pass

try:
    import informixdb

    # LOGGER.warning('Informix support is experimental')
    DRIVERS["informixdb"] = informixdb
except ImportError:
    pass

try:
    import sapdb

    DRIVERS["sapdb"] = sapdb
    # LOGGER.warning('SAPDB support is experimental')
except ImportError:
    pass

try:
    import cubriddb

    DRIVERS["cubriddb"] = cubriddb
    # LOGGER.warning('Cubrid support is experimental')
except ImportError:
    pass

try:
    import java.sql
    from com.ziclix.python.sql import zxJDBC

    # Try sqlite jdbc driver from http://www.zentus.com/sqlitejdbc/
    from org.sqlite import JDBC  # required by java.sql; ensure we have it

    zxJDBC_sqlite = java.sql.DriverManager
    DRIVERS["zxJDBC"] = zxJDBC
    # DRIVERS.append('SQLite(zxJDBC)')
    # LOGGER.warning('zxJDBC support is experimental')
    is_jdbc = True
except ImportError:
    is_jdbc = False

try:
    import couchdb

    DRIVERS["couchdb"] = couchdb
except ImportError:
    couchdb = None

try:
    import pymongo

    DRIVERS["pymongo"] = pymongo
except:
    pass

try:
    import imaplib

    DRIVERS["imaplib"] = imaplib
except:
    pass

try:
    import pytds

    DRIVERS["pytds"] = pytds
except:
    pass

try:
    import pymssql

    DRIVERS["pymssql"] = pymssql
except:
    pass

try:
    import mssql_python

    DRIVERS["mssql-python"] = mssql_python
except:
    pass



import datetime
import functools

from .. import Field
from ..validators import *


class Tags:
    def __init__(self, table, name="default", tag_table=None):
        self.name = name
        self.table = table
        self.tag_table = tag_table or self._make_tag_table()

    def _make_tag_table(self):
        db = self.table._db
        tag_table = db.define_table(
            self.table._tablename + "_tag_" + self.name,
            Field("tagpath"),
            Field("record_id", self.table),
        )
        db.commit()
        return tag_table

    def get(self, record_id):
        tag_table = self.tag_table
        db = tag_table._db
        rows = db(tag_table.record_id == record_id).select(tag_table.tagpath)
        return [row.tagpath.strip("/") for row in rows]

    def add(self, record_id, tags):
        tag_table = self.tag_table
        db = tag_table._db
        if not isinstance(tags, list):
            tags = [tags]
        for tag in tags:
            path = "/%s/" % tag.strip("/")
            if not db(
                (tag_table.record_id == record_id) & (tag_table.tagpath == path)
            ).count():
                tag_table.insert(record_id=record_id, tagpath=path)

    def remove(self, record_id, tags):
        tag_table = self.tag_table
        db = tag_table._db
        if not isinstance(tags, list):
            tags = [tags]
        paths = ["/%s/" % tag.strip("/") for tag in tags]
        db(
            (tag_table.record_id == record_id) & (tag_table.tagpath.belongs(paths))
        ).delete()

    def find(self, tags, mode="and"):
        table = self.table
        tag_table = self.tag_table
        db = tag_table._db
        queries = []
        if not isinstance(tags, list):
            tags = [tags]
        for tag in tags:
            path = "/%s/" % tag.strip("/")
            subquery = db(tag_table.tagpath.startswith(path))._select(
                tag_table.record_id
            )
            queries.append(table.id.belongs(subquery))
        func = lambda a, b: (a & b) if mode == "and" else (a | b)
        return functools.reduce(func, queries)



import re

from .._compat import integer_types, long
from . import adapters
from .base import SQLAdapter


@adapters.register_for("sapdb")
class SAPDB(SQLAdapter):
    dbengine = "sapdb"
    drivers = ("sapdb",)

    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]+|\[[^\]]+\])/(?P<db>[^?]+)$"
    )

    def _initialize_(self):
        super(SAPDB, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        user = self.credential_decoder(m.group("user"))
        password = self.credential_decoder(m.group("password"))
        if password is None:
            password = ""
        host = m.group("host")
        db = m.group("db")
        self.driver_args.update(user=user, password=password, database=db, host=host)

    def connector(self):
        self.driver.connect(**self.driver_args)

    def lastrowid(self, table):
        self.execute("select %s.NEXTVAL from dual" % table._sequence_name)
        return int(self.cursor.fetchone()[0])

    def create_sequence_and_triggers(self, query, table, **args):
        self.execute("CREATE SEQUENCE %s;" % table._sequence_name)
        self.execute(
            "ALTER TABLE %s ALTER COLUMN %s SET DEFAULT NEXTVAL('%s');"
            % (table._rname, table._id._rname, table._sequence_name)
        )
        self.execute(query)



import hashlib
import os
import sys

PY2 = sys.version_info[0] == 2

_identity = lambda x: x

if PY2:
    from email import Charset, Encoders
    from email.Charset import QP as charset_QP
    from email.Charset import add_charset
    from email.Header import Header
    from email.MIMEBase import MIMEBase
    from email.MIMEMultipart import MIMEMultipart
    from email.MIMEText import MIMEText
    from string import maketrans
    from types import ClassType
    from urllib import quote as urllib_quote
    from urllib import quote_plus as urllib_quote_plus
    from urllib import unquote
    from urllib import unquote as urllib_unquote
    from urllib import urlencode

    import __builtin__ as builtin
    import ConfigParser as configparser
    import Cookie
    import cookielib
    import copy_reg as copyreg
    import cPickle as pickle
    import Queue
    import thread
    import urllib2
    import urlparse
    from cStringIO import StringIO
    from htmlentitydefs import entitydefs, name2codepoint
    from HTMLParser import HTMLParser
    from urllib2 import urlopen
    from xmlrpclib import ProtocolError

    from .contrib import ipaddress

    BytesIO = StringIO
    reduce = reduce
    reload = reload
    hashlib_md5 = hashlib.md5
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()
    integer_types = (int, long)
    string_types = (str, unicode)
    text_type = unicode
    basestring = basestring
    xrange = xrange
    long = long
    unichr = unichr
    unicodeT = unicode

    def implements_bool(cls):
        cls.__nonzero__ = cls.__bool__
        del cls.__bool__
        return cls

    def implements_iterator(cls):
        cls.next = cls.__next__
        del cls.__next__
        return cls

    def to_bytes(obj, charset="utf-8", errors="strict"):
        if obj is None:
            return None
        if isinstance(obj, (bytes, bytearray, buffer)):
            return bytes(obj)
        if isinstance(obj, unicode):
            return obj.encode(charset, errors)
        raise TypeError("Expected bytes")

    def to_native(obj, charset="utf8", errors="strict"):
        if obj is None or isinstance(obj, str):
            return obj
        return obj.encode(charset, errors)

else:
    import _thread as thread
    import builtins as builtin
    import configparser
    import copyreg
    import html  # warning, this is the python3 module and not the web2py html module
    import ipaddress
    import pickle
    import queue as Queue
    from email import encoders as Encoders
    from email.charset import QP as charset_QP
    from email.charset import Charset, add_charset
    from email.header import Header
    from email.mime.base import MIMEBase
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText
    from functools import reduce
    from html.entities import entitydefs, name2codepoint
    from html.parser import HTMLParser
    from http import cookiejar as cookielib
    from http import cookies as Cookie
    from importlib import reload
    from io import BytesIO, StringIO
    from urllib import parse as urlparse
    from urllib import request as urllib2
    from urllib.parse import quote as urllib_quote
    from urllib.parse import quote_plus as urllib_quote_plus
    from urllib.parse import unquote
    from urllib.parse import unquote as urllib_unquote
    from urllib.parse import urlencode
    from urllib.request import urlopen
    from xmlrpc.client import ProtocolError

    hashlib_md5 = lambda s: hashlib.md5(bytes(s, "utf8"))
    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())
    integer_types = (int,)
    string_types = (str,)
    text_type = str
    basestring = str
    xrange = range
    long = int
    unichr = chr
    unicodeT = str
    maketrans = str.maketrans
    ClassType = type

    implements_iterator = _identity
    implements_bool = _identity

    def to_bytes(obj, charset="utf-8", errors="strict"):
        if obj is None:
            return None
        if isinstance(obj, (bytes, bytearray, memoryview)):
            return bytes(obj)
        if isinstance(obj, str):
            return obj.encode(charset, errors)
        raise TypeError("Expected bytes")

    def to_native(obj, charset="utf8", errors="strict"):
        if obj is None or isinstance(obj, str):
            return obj
        return obj.decode(charset, errors)


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""

    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)

    return metaclass("temporary_class", None, {})


def to_unicode(obj, charset="utf-8", errors="strict"):
    if obj is None:
        return None
    if not hasattr(obj, "decode") or not callable(obj.decode):
        return text_type(obj)
    return obj.decode(charset, errors)


# shortcuts
pjoin = os.path.join
exists = os.path.exists



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Unit tests for validators.py"""

import datetime
import decimal
import re
import unittest
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    import cgi

from pydal import DAL, Field
from pydal._compat import PY2, to_bytes
from pydal.validators import *
from pydal.validators import UTC, Validator, options_sorter


class TestValidators(unittest.TestCase):
    def myassertRegex(self, *args, **kwargs):
        if PY2:
            return getattr(self, "assertRegexpMatches")(*args, **kwargs)
        return getattr(self, "assertRegex")(*args, **kwargs)

    def test_MISC(self):
        """Test miscelaneous utility functions and some general behavior guarantees"""

        self.assertEqual(options_sorter(("a", "a"), ("a", "a")), -1)
        self.assertEqual(options_sorter(("A", "A"), ("a", "a")), -1)
        self.assertEqual(options_sorter(("b", "b"), ("a", "a")), 1)
        utc = UTC()
        dt = datetime.datetime.now()
        self.assertEqual(utc.utcoffset(dt), UTC.ZERO)
        self.assertEqual(utc.dst(dt), UTC.ZERO)
        self.assertEqual(utc.tzname(dt), "UTC")

    def test_IS_MATCH(self):
        rtn = IS_MATCH(".+")("hello")
        self.assertEqual(rtn, ("hello", None))
        rtn = IS_MATCH("hell")("hello")
        self.assertEqual(rtn, ("hello", None))
        rtn = IS_MATCH("hell.*", strict=False)("hello")
        self.assertEqual(rtn, ("hello", None))
        rtn = IS_MATCH("hello")("shello")
        self.assertEqual(rtn, ("shello", "Invalid expression"))
        rtn = IS_MATCH("hello", search=True)("shello")
        self.assertEqual(rtn, ("shello", None))
        rtn = IS_MATCH("hello", search=True, strict=False)("shellox")
        self.assertEqual(rtn, ("shellox", None))
        rtn = IS_MATCH(".*hello.*", search=True, strict=False)("shellox")
        self.assertEqual(rtn, ("shellox", None))
        rtn = IS_MATCH(".+")("")
        self.assertEqual(rtn, ("", "Invalid expression"))
        rtn = IS_MATCH("hell", strict=True)("hellas")
        self.assertEqual(rtn, ("hellas", "Invalid expression"))
        rtn = IS_MATCH("hell$", strict=True)("hellas")
        self.assertEqual(rtn, ("hellas", "Invalid expression"))
        rtn = IS_MATCH("^.hell$", strict=True)("shell")
        self.assertEqual(rtn, ("shell", None))
        rtn = IS_MATCH("hell", is_unicode=True)("àòè")
        if PY2:
            self.assertEqual(rtn, ("\xc3\xa0\xc3\xb2\xc3\xa8", "Invalid expression"))
        else:
            self.assertEqual(rtn, ("àòè", "Invalid expression"))
        rtn = IS_MATCH("hell", is_unicode=True)("hell")
        self.assertEqual(rtn, ("hell", None))
        rtn = IS_MATCH("hell", is_unicode=True)("hell")
        self.assertEqual(rtn, ("hell", None))
        # regr test for #1044
        rtn = IS_MATCH("hello")("\xff")
        self.assertEqual(rtn, ("\xff", "Invalid expression"))

    def test_IS_EQUAL_TO(self):
        rtn = IS_EQUAL_TO("aaa")("aaa")
        self.assertEqual(rtn, ("aaa", None))
        rtn = IS_EQUAL_TO("aaa")("aab")
        self.assertEqual(rtn, ("aab", "No match"))

    def test_IS_EXPR(self):
        rtn = IS_EXPR("int(value) < 2")("1")
        self.assertEqual(rtn, ("1", None))
        rtn = IS_EXPR("int(value) < 2")("2")
        self.assertEqual(rtn, ("2", "Invalid expression"))
        rtn = IS_EXPR(lambda value: int(value))("1")
        self.assertEqual(rtn, ("1", 1))
        rtn = IS_EXPR(lambda value: int(value) < 2 and "invalid" or None)("2")
        self.assertEqual(rtn, ("2", None))

    def test_IS_LENGTH(self):
        rtn = IS_LENGTH()("")
        self.assertEqual(rtn, ("", None))
        rtn = IS_LENGTH()("1234567890")
        self.assertEqual(rtn, ("1234567890", None))
        rtn = IS_LENGTH(maxsize=5, minsize=0)("1234567890")  # too long
        self.assertEqual(rtn, ("1234567890", "Enter from 0 to 5 characters"))
        rtn = IS_LENGTH(maxsize=50, minsize=20)("1234567890")  # too short
        self.assertEqual(rtn, ("1234567890", "Enter from 20 to 50 characters"))
        rtn = IS_LENGTH()(None)
        self.assertEqual(rtn, (None, None))
        rtn = IS_LENGTH(minsize=0)(None)
        self.assertEqual(rtn, (None, None))
        rtn = IS_LENGTH(minsize=1)(None)
        self.assertEqual(rtn, (None, "Enter from 1 to 255 characters"))
        rtn = IS_LENGTH(minsize=1)([])
        self.assertEqual(rtn, ([], "Enter from 1 to 255 characters"))
        rtn = IS_LENGTH(minsize=1)([1, 2])
        self.assertEqual(rtn, ([1, 2], None))
        rtn = IS_LENGTH(minsize=1)([1])
        self.assertEqual(rtn, ([1], None))
        # test non utf-8 str
        cpstr = "lálá".encode("cp1252")
        rtn = IS_LENGTH(minsize=4)(cpstr)
        self.assertEqual(rtn, (cpstr, None))
        rtn = IS_LENGTH(maxsize=4)(cpstr)
        self.assertEqual(rtn, (cpstr, None))
        rtn = IS_LENGTH(minsize=0, maxsize=3)(cpstr)
        self.assertEqual(rtn, (cpstr, "Enter from 0 to 3 characters"))
        # test unicode
        rtn = IS_LENGTH(2)("°2")
        if PY2:
            self.assertEqual(rtn, ("\xc2\xb02", None))
        else:
            self.assertEqual(rtn, ("°2", None))
        rtn = IS_LENGTH(2)("°12")
        if PY2:
            self.assertEqual(rtn, ("\xb012", "Enter from 0 to 2 characters"))
        else:
            self.assertEqual(rtn, ("°12", "Enter from 0 to 2 characters"))
        # test automatic str()
        rtn = IS_LENGTH(minsize=1)(1)
        self.assertEqual(rtn, ("1", None))
        rtn = IS_LENGTH(minsize=2)(1)
        self.assertEqual(rtn, (1, "Enter from 2 to 255 characters"))
        # test FieldStorage
        from io import BytesIO

        a = cgi.FieldStorage()
        a.file = BytesIO(b"abc")
        rtn = IS_LENGTH(minsize=4)(a)
        self.assertEqual(rtn, (a, "Enter from 4 to 255 characters"))
        urlencode_data = b"key2=value2x&key3=value3&key4=value4"
        urlencode_environ = {
            "CONTENT_LENGTH": str(len(urlencode_data)),
            "CONTENT_TYPE": "application/x-www-form-urlencoded",
            "QUERY_STRING": "key1=value1&key2=value2y",
            "REQUEST_METHOD": "POST",
        }
        fake_stdin = BytesIO(urlencode_data)
        fake_stdin.seek(0)
        a = cgi.FieldStorage(fp=fake_stdin, environ=urlencode_environ)
        rtn = IS_LENGTH(minsize=6)(a)
        self.assertEqual(rtn, (a, "Enter from 6 to 255 characters"))
        a = cgi.FieldStorage()
        rtn = IS_LENGTH(minsize=6)(a)
        self.assertEqual(rtn, (a, "Enter from 6 to 255 characters"))
        rtn = IS_LENGTH(6)(a)
        self.assertEqual(rtn, (a, None))

    def test_IS_JSON(self):
        rtn = IS_JSON()('{"a": 100}')
        self.assertEqual(rtn, ({"a": 100}, None))
        rtn = IS_JSON()("spam1234")
        self.assertEqual(rtn, ("spam1234", "Invalid json"))
        rtn = IS_JSON(native_json=True)('{"a": 100}')
        self.assertEqual(rtn, ('{"a": 100}', None))
        rtn = IS_JSON().formatter(None)
        self.assertEqual(rtn, "null")
        rtn = IS_JSON().formatter({"a": 100})
        self.assertEqual(rtn, '{"a": 100}')
        rtn = IS_JSON(native_json=True).formatter({"a": 100})
        self.assertEqual(rtn, {"a": 100})

    def test_IS_IN_SET(self):
        rtn = IS_IN_SET(["max", "john"])("max")
        self.assertEqual(rtn, ("max", None))
        rtn = IS_IN_SET(["max", "john"])("massimo")
        self.assertEqual(rtn, ("massimo", "Value not allowed"))
        rtn = IS_IN_SET(["max", "john"], multiple=True)(("max", "john"))
        self.assertEqual(rtn, (("max", "john"), None))
        # order reversed should still be fine
        rtn = IS_IN_SET(["max", "john"], multiple=True)(("john", "max"))
        self.assertEqual(rtn, (("john", "max"), None))
        rtn = IS_IN_SET(["max", "john"], multiple=True)(("bill", "john"))
        self.assertEqual(rtn, (("bill", "john"), "Value not allowed"))
        rtn = IS_IN_SET(("id1", "id2"), ["first label", "second label"])(
            "id1"
        )  # Traditional way
        self.assertEqual(rtn, ("id1", None))
        rtn = IS_IN_SET({"id1": "first label", "id2": "second label"})("id1")
        self.assertEqual(rtn, ("id1", None))
        rtn = IS_IN_SET(["id1", "id2"], error_message="oops", multiple=True)(None)
        self.assertEqual(rtn, ([], None))
        rtn = IS_IN_SET(["id1", "id2"], error_message="oops", multiple=True)("")
        self.assertEqual(rtn, ([], None))
        rtn = IS_IN_SET(["id1", "id2"], error_message="oops", multiple=True)("id1")
        self.assertEqual(rtn, (["id1"], None))
        rtn = IS_IN_SET(["id1", "id2"], error_message="oops", multiple=(1, 2))(None)
        self.assertEqual(rtn, (None, "oops"))
        import itertools

        rtn = IS_IN_SET(itertools.chain(["1", "3", "5"], ["2", "4", "6"]))("1")
        self.assertEqual(rtn, ("1", None))
        rtn = IS_IN_SET([("id1", "first label"), ("id2", "second label")])(
            "id1"
        )  # Redundant way
        self.assertEqual(rtn, ("id1", None))
        rtn = IS_IN_SET([("id1", "first label"), ("id2", "second label")]).options(
            zero=False
        )
        self.assertEqual(rtn, [("id1", "first label"), ("id2", "second label")])
        rtn = IS_IN_SET(["id1", "id2"]).options(zero=False)
        self.assertEqual(rtn, [("id1", "id1"), ("id2", "id2")])
        rtn = IS_IN_SET(["id2", "id1"], sort=True).options(zero=False)
        self.assertEqual(rtn, [("id1", "id1"), ("id2", "id2")])

        # sets can be lazy
        theset = []
        validator = IS_IN_SET(lambda: theset)
        rtn = validator("max")
        self.assertEqual(rtn, ("max", "Value not allowed"))
        theset.append("max")
        rtn = validator("max")
        self.assertEqual(rtn, ("max", None))

    def test_IS_IN_DB(self):
        db = DAL("sqlite:memory")
        db.define_table("person", Field("name"))
        george_id = db.person.insert(name="george")
        costanza_id = db.person.insert(name="costanza")
        rtn = IS_IN_DB(db, "person.id", "%(name)s")(george_id)
        self.assertEqual(rtn, (george_id, None))
        rtn = IS_IN_DB(db, "person.name", "%(name)s")("george")
        self.assertEqual(rtn, ("george", None))
        rtn = IS_IN_DB(db, db.person, "%(name)s")(george_id)
        self.assertEqual(rtn, (george_id, None))
        rtn = IS_IN_DB(db(db.person.id > 0), db.person, "%(name)s")(george_id)
        self.assertEqual(rtn, (george_id, None))
        rtn = IS_IN_DB(db, "person.id", "%(name)s", error_message="oops")(
            george_id + costanza_id
        )
        self.assertEqual(rtn, (george_id + costanza_id, "oops"))
        rtn = IS_IN_DB(db, db.person.id, "%(name)s")(george_id)
        self.assertEqual(rtn, (george_id, None))
        rtn = IS_IN_DB(db, db.person.id, "%(name)s", error_message="oops")(
            george_id + costanza_id
        )
        self.assertEqual(rtn, (george_id + costanza_id, "oops"))
        rtn = IS_IN_DB(db, "person.id", "%(name)s", multiple=True)(
            [george_id, costanza_id]
        )
        self.assertEqual(rtn, ([george_id, costanza_id], None))
        rtn = IS_IN_DB(
            db, "person.id", "%(name)s", multiple=True, error_message="oops"
        )("I'm not even an id")
        self.assertEqual(rtn, ("I'm not even an id", "oops"))
        rtn = IS_IN_DB(db, "person.id", "%(name)s", multiple=True, delimiter=",")(
            "%d,%d" % (george_id, costanza_id)
        )
        self.assertEqual(rtn, (("%d,%d" % (george_id, costanza_id)).split(","), None))
        rtn = IS_IN_DB(db, "person.id", "%(name)s", multiple=(1, 3), delimiter=",")(
            "%d,%d" % (george_id, costanza_id)
        )
        self.assertEqual(rtn, (("%d,%d" % (george_id, costanza_id)).split(","), None))
        rtn = IS_IN_DB(
            db,
            "person.id",
            "%(name)s",
            multiple=(1, 2),
            delimiter=",",
            error_message="oops",
        )("%d,%d" % (george_id, costanza_id))
        self.assertEqual(rtn, (("%d,%d" % (george_id, costanza_id)), "oops"))
        rtn = IS_IN_DB(db, db.person.id, "%(name)s", error_message="oops").options(
            zero=False
        )
        self.assertEqual(
            sorted(rtn),
            [("%d" % george_id, "george"), ("%d" % costanza_id, "costanza")],
        )
        rtn = IS_IN_DB(
            db, db.person.id, db.person.name, error_message="oops", sort=True
        ).options(zero=True)
        self.assertEqual(
            rtn,
            [("", ""), ("%d" % costanza_id, "costanza"), ("%d" % george_id, "george")],
        )
        # Test None
        rtn = IS_IN_DB(db, "person.id", "%(name)s", error_message="oops")(None)
        self.assertEqual(rtn, (None, "oops"))
        rtn = IS_IN_DB(db, "person.name", "%(name)s", error_message="oops")(None)
        self.assertEqual(rtn, (None, "oops"))
        # Test using the set it made for options
        vldtr = IS_IN_DB(db, "person.name", "%(name)s", error_message="oops")
        vldtr.options()
        rtn = vldtr("george")
        self.assertEqual(rtn, ("george", None))
        rtn = vldtr("jerry")
        self.assertEqual(rtn, ("jerry", "oops"))
        vldtr = IS_IN_DB(
            db, "person.name", "%(name)s", error_message="oops", multiple=True
        )
        vldtr.options()
        rtn = vldtr(["george", "costanza"])
        self.assertEqual(rtn, (["george", "costanza"], None))
        # Test it works with self reference
        db.define_table(
            "category",
            Field(
                "parent_id",
                "reference category",
                requires=IS_EMPTY_OR(IS_IN_DB(db, "category.id", "%(name)s")),
            ),
            Field("name"),
        )
        ret = db.category.validate_and_insert(name="seinfeld")
        self.assertFalse(ret.get("errors"))
        ret = db.category.validate_and_insert(
            name="characters", parent_id=ret.get("id")
        )
        self.assertFalse(ret.get("errors"))
        rtn = IS_IN_DB(db, "category.id", "%(name)s")(ret.get("id"))
        self.assertEqual(rtn, (ret.get("id"), None))
        # Test _and
        vldtr = IS_IN_DB(
            db,
            "person.name",
            "%(name)s",
            error_message="oops",
            _and=IS_LENGTH(maxsize=7, error_message="bad"),
        )
        rtn = vldtr("george")
        self.assertEqual(rtn, ("george", None))
        rtn = vldtr("costanza")
        self.assertEqual(rtn, ("costanza", "bad"))
        rtn = vldtr("jerry")
        self.assertEqual(rtn, ("jerry", "oops"))
        vldtr.options()  # test theset with _and
        rtn = vldtr("jerry")
        self.assertEqual(rtn, ("jerry", "oops"))
        # Test auto_add
        rtn = IS_IN_DB(db, "person.id", "%(name)s", error_message="oops")("jerry")
        self.assertEqual(rtn, ("jerry", "oops"))
        rtn = IS_IN_DB(db, "person.id", "%(name)s", auto_add=True)("jerry")
        self.assertEqual(rtn, (3, None))
        # Test it works with reference table
        db.define_table(
            "ref_table", Field("name"), Field("person_id", "reference person")
        )
        ret = db.ref_table.validate_and_insert(name="test reference table")
        self.assertTrue(not ret.get("errors"))
        ret = db.ref_table.validate_and_insert(
            name="test reference table", person_id=george_id
        )
        self.assertTrue(not ret.get("errors"))
        rtn = IS_IN_DB(db, "ref_table.person_id", "%(name)s")(george_id)
        self.assertEqual(rtn, (george_id, None))
        # Test it works with reference table.field and keyed table
        db.define_table("person_keyed", Field("name"), primarykey=["name"])
        db.person_keyed.insert(name="george")
        db.person_keyed.insert(name="costanza")
        rtn = IS_IN_DB(db, "person_keyed.name")("george")
        self.assertEqual(rtn, ("george", None))
        db.define_table(
            "ref_table_field",
            Field("name"),
            Field("person_name", "reference person_keyed.name", requires=[]),
        )
        ret = db.ref_table_field.validate_and_insert(name="test reference table.field")
        self.assertTrue(not ret.get("errors"))
        ret = db.ref_table_field.validate_and_insert(
            name="test reference table.field", person_name="george"
        )
        self.assertTrue(not ret.get("errors"))
        vldtr = IS_IN_DB(db, "ref_table_field.person_name", "%(name)s")
        vldtr.options()
        rtn = vldtr("george")
        self.assertEqual(rtn, ("george", None))
        # Test it works with list:reference table
        db.define_table(
            "list_ref_table",
            Field("name"),
            Field("person_list", "list:reference person"),
        )
        self.assertEqual(db.list_ref_table.person_list.referenced_table(), db.person)
        ret = db.list_ref_table.validate_and_insert(name="test list:reference table")
        self.assertTrue(not ret.get("errors"))
        ret = db.list_ref_table.validate_and_insert(
            name="test list:reference table", person_list=[george_id, costanza_id]
        )
        self.assertTrue(not ret.get("errors"))
        vldtr = IS_IN_DB(db, "list_ref_table.person_list")
        vldtr.options()
        rtn = vldtr([george_id, costanza_id])
        self.assertEqual(rtn, ([george_id, costanza_id], None))
        db.person.drop()
        db.category.drop()
        db.person_keyed.drop()
        db.ref_table.drop()
        db.ref_table_field.drop()
        db.list_ref_table.drop()

    def test_IS_IN_DB_options(self):
        db = DAL("sqlite:memory")
        db.define_table("thing", Field("name"))
        db.thing.insert(name="red")
        db.thing.insert(name="green")
        assert IS_IN_DB(db, "thing").options() == [("", ""), ("1", "1"), ("2", "2")]
        # setting a format for the validator
        db.thing._format = lambda i: i.name
        assert IS_IN_DB(db, "thing").options() == [
            ("", ""),
            ("1", "red"),
            ("2", "green"),
        ]
        assert IS_IN_DB(db, "thing", zero=None).options() == [
            ("1", "red"),
            ("2", "green"),
        ]
        assert IS_IN_DB(db, "thing", zero="x").options() == [
            ("", "x"),
            ("1", "red"),
            ("2", "green"),
        ]

    def test_IS_NOT_IN_DB(self):
        db = DAL("sqlite:memory")
        db.define_table("person", Field("name"), Field("nickname"))
        db.person.insert(name="george")
        costanza_id = db.person.insert(name="costanza", nickname="T Bone")
        rtn = IS_NOT_IN_DB(db, "person.name", error_message="oops")("george")
        self.assertEqual(rtn, ("george", "oops"))
        rtn = IS_NOT_IN_DB(
            db, "person.name", error_message="oops", allowed_override=["george"]
        )("george")
        self.assertEqual(rtn, ("george", None))
        rtn = IS_NOT_IN_DB(db, "person.name", error_message="oops")("  ")
        self.assertEqual(rtn, ("  ", "oops"))
        rtn = IS_NOT_IN_DB(db, "person.name")("jerry")
        self.assertEqual(rtn, ("jerry", None))
        rtn = IS_NOT_IN_DB(db, "person.name")("jerry")
        self.assertEqual(rtn, ("jerry", None))
        rtn = IS_NOT_IN_DB(db(db.person.id > 0), "person.name")("jerry")
        self.assertEqual(rtn, ("jerry", None))
        rtn = IS_NOT_IN_DB(db, db.person, error_message="oops")(1)
        self.assertEqual(rtn, (1, "oops"))
        vldtr = IS_NOT_IN_DB(db, "person.name", error_message="oops")
        vldtr.set_self_id(costanza_id)
        rtn = vldtr("george")
        self.assertEqual(rtn, ("george", "oops"))
        rtn = vldtr("costanza")
        self.assertEqual(rtn, ("costanza", None))

        db.person.drop()

    def test_IS_INT_IN_RANGE(self):
        rtn = IS_INT_IN_RANGE(1, 5)("4")
        self.assertEqual(rtn, (4, None))
        rtn = IS_INT_IN_RANGE(1, 5)(4)
        self.assertEqual(rtn, (4, None))
        rtn = IS_INT_IN_RANGE(1, 5)(1)
        self.assertEqual(rtn, (1, None))
        rtn = IS_INT_IN_RANGE(1, 5)(5)
        self.assertEqual(rtn, (5, "Enter an integer between 1 and 4"))
        rtn = IS_INT_IN_RANGE(1, 5)(5)
        self.assertEqual(rtn, (5, "Enter an integer between 1 and 4"))
        rtn = IS_INT_IN_RANGE(1, 5)(3.5)
        self.assertEqual(rtn, (3.5, "Enter an integer between 1 and 4"))
        rtn = IS_INT_IN_RANGE(None, 5)("4")
        self.assertEqual(rtn, (4, None))
        rtn = IS_INT_IN_RANGE(None, 5)("6")
        self.assertEqual(rtn, ("6", "Enter an integer less than or equal to 4"))
        rtn = IS_INT_IN_RANGE(1, None)("4")
        self.assertEqual(rtn, (4, None))
        rtn = IS_INT_IN_RANGE(1, None)("0")
        self.assertEqual(rtn, ("0", "Enter an integer greater than or equal to 1"))
        rtn = IS_INT_IN_RANGE()(6)
        self.assertEqual(rtn, (6, None))
        rtn = IS_INT_IN_RANGE()("abc")
        self.assertEqual(rtn, ("abc", "Enter an integer"))

    def test_IS_FLOAT_IN_RANGE(self):
        # with None
        rtn = IS_FLOAT_IN_RANGE(1, 5)(None)
        self.assertEqual(rtn, (None, "Enter a number between 1 and 5"))
        rtn = IS_FLOAT_IN_RANGE(1, 5)("4")
        self.assertEqual(rtn, (4.0, None))
        rtn = IS_FLOAT_IN_RANGE(1, 5)(4)
        self.assertEqual(rtn, (4.0, None))
        rtn = IS_FLOAT_IN_RANGE(1, 5)(1)
        self.assertEqual(rtn, (1.0, None))
        rtn = IS_FLOAT_IN_RANGE(1, 5)(5.25)
        self.assertEqual(rtn, (5.25, "Enter a number between 1 and 5"))
        rtn = IS_FLOAT_IN_RANGE(1, 5)(6.0)
        self.assertEqual(rtn, (6.0, "Enter a number between 1 and 5"))
        rtn = IS_FLOAT_IN_RANGE(1, 5)(3.5)
        self.assertEqual(rtn, (3.5, None))
        rtn = IS_FLOAT_IN_RANGE(1, None)(3.5)
        self.assertEqual(rtn, (3.5, None))
        rtn = IS_FLOAT_IN_RANGE(None, 5)(3.5)
        self.assertEqual(rtn, (3.5, None))
        rtn = IS_FLOAT_IN_RANGE(1, None)(0.5)
        self.assertEqual(rtn, (0.5, "Enter a number greater than or equal to 1"))
        rtn = IS_FLOAT_IN_RANGE(None, 5)(6.5)
        self.assertEqual(rtn, (6.5, "Enter a number less than or equal to 5"))
        rtn = IS_FLOAT_IN_RANGE()(6.5)
        self.assertEqual(rtn, (6.5, None))
        rtn = IS_FLOAT_IN_RANGE()("abc")
        self.assertEqual(rtn, ("abc", "Enter a number"))
        rtn = IS_FLOAT_IN_RANGE()("6,5")
        self.assertEqual(rtn, ("6,5", "Enter a number"))
        rtn = IS_FLOAT_IN_RANGE(dot=",")("6.5")
        self.assertEqual(rtn, (6.5, None))
        # With .formatter(None)
        rtn = IS_FLOAT_IN_RANGE(dot=",").formatter(None)
        self.assertEqual(rtn, None)
        rtn = IS_FLOAT_IN_RANGE(dot=",").formatter(0.25)
        self.assertEqual(rtn, "0,25")
        # To trigger str2dec "if not '.' in s:" line
        rtn = IS_FLOAT_IN_RANGE(dot=",").formatter(1)
        self.assertEqual(rtn, "1,00")

    def test_IS_DECIMAL_IN_RANGE(self):
        # with None
        rtn = IS_DECIMAL_IN_RANGE(1, 5)(None)
        self.assertEqual(rtn, (None, "Enter a number between 1 and 5"))
        rtn = IS_DECIMAL_IN_RANGE(1, 5)("4")
        self.assertEqual(rtn, (decimal.Decimal("4"), None))
        rtn = IS_DECIMAL_IN_RANGE(1, 5)(4)
        self.assertEqual(rtn, (decimal.Decimal("4"), None))
        rtn = IS_DECIMAL_IN_RANGE(1, 5)(1)
        self.assertEqual(rtn, (decimal.Decimal("1"), None))
        rtn = IS_DECIMAL_IN_RANGE(1, 5)(5.25)
        self.assertEqual(rtn, (5.25, "Enter a number between 1 and 5"))
        rtn = IS_DECIMAL_IN_RANGE(5.25, 6)(5.25)
        self.assertEqual(rtn, (decimal.Decimal("5.25"), None))
        rtn = IS_DECIMAL_IN_RANGE(5.25, 6)("5.25")
        self.assertEqual(rtn, (decimal.Decimal("5.25"), None))
        rtn = IS_DECIMAL_IN_RANGE(1, 5)(6.0)
        self.assertEqual(rtn, (6.0, "Enter a number between 1 and 5"))
        rtn = IS_DECIMAL_IN_RANGE(1, 5)(3.5)
        self.assertEqual(rtn, (decimal.Decimal("3.5"), None))
        rtn = IS_DECIMAL_IN_RANGE(1.5, 5.5)(3.5)
        self.assertEqual(rtn, (decimal.Decimal("3.5"), None))
        rtn = IS_DECIMAL_IN_RANGE(1.5, 5.5)(6.5)
        self.assertEqual(rtn, (6.5, "Enter a number between 1.5 and 5.5"))
        rtn = IS_DECIMAL_IN_RANGE(1.5, None)(6.5)
        self.assertEqual(rtn, (decimal.Decimal("6.5"), None))
        rtn = IS_DECIMAL_IN_RANGE(1.5, None)(0.5)
        self.assertEqual(rtn, (0.5, "Enter a number greater than or equal to 1.5"))
        rtn = IS_DECIMAL_IN_RANGE(None, 5.5)(4.5)
        self.assertEqual(rtn, (decimal.Decimal("4.5"), None))
        rtn = IS_DECIMAL_IN_RANGE(None, 5.5)(6.5)
        self.assertEqual(rtn, (6.5, "Enter a number less than or equal to 5.5"))
        rtn = IS_DECIMAL_IN_RANGE()(6.5)
        self.assertEqual(rtn, (decimal.Decimal("6.5"), None))
        rtn = IS_DECIMAL_IN_RANGE(0, 99)(123.123)
        self.assertEqual(rtn, (123.123, "Enter a number between 0 and 99"))
        rtn = IS_DECIMAL_IN_RANGE(0, 99)("123.123")
        self.assertEqual(rtn, ("123.123", "Enter a number between 0 and 99"))
        rtn = IS_DECIMAL_IN_RANGE(0, 99)("12.34")
        self.assertEqual(rtn, (decimal.Decimal("12.34"), None))
        rtn = IS_DECIMAL_IN_RANGE()("abc")
        self.assertEqual(rtn, ("abc", "Enter a number"))
        rtn = IS_DECIMAL_IN_RANGE()("6,5")
        self.assertEqual(rtn, ("6,5", "Enter a number"))
        rtn = IS_DECIMAL_IN_RANGE(dot=",")("6.5")
        self.assertEqual(rtn, (decimal.Decimal("6.5"), None))
        rtn = IS_DECIMAL_IN_RANGE(1, 5)(decimal.Decimal("4"))
        self.assertEqual(rtn, (decimal.Decimal("4"), None))
        # With .formatter(None)
        rtn = IS_DECIMAL_IN_RANGE(dot=",").formatter(None)
        self.assertEqual(rtn, None)
        rtn = IS_DECIMAL_IN_RANGE(dot=",").formatter(0.25)
        self.assertEqual(rtn, "0,25")

    def test_IS_NOT_EMPTY(self):
        rtn = IS_NOT_EMPTY()(1)
        self.assertEqual(rtn, (1, None))
        rtn = IS_NOT_EMPTY()(0)
        self.assertEqual(rtn, (0, None))
        rtn = IS_NOT_EMPTY()("x")
        self.assertEqual(rtn, ("x", None))
        rtn = IS_NOT_EMPTY()(" x ")
        self.assertEqual(rtn, (" x ", None))
        rtn = IS_NOT_EMPTY()(None)
        self.assertEqual(rtn, (None, "Enter a value"))
        rtn = IS_NOT_EMPTY()("")
        self.assertEqual(rtn, ("", "Enter a value"))
        rtn = IS_NOT_EMPTY()(b"")
        self.assertEqual(rtn, (b"", "Enter a value"))
        rtn = IS_NOT_EMPTY()("  ")
        self.assertEqual(rtn, ("  ", "Enter a value"))
        rtn = IS_NOT_EMPTY()(" \n\t")
        self.assertEqual(rtn, (" \n\t", "Enter a value"))
        rtn = IS_NOT_EMPTY()([])
        self.assertEqual(rtn, ([], "Enter a value"))
        rtn = IS_NOT_EMPTY(empty_regex="def")("def")
        self.assertEqual(rtn, ("def", "Enter a value"))
        rtn = IS_NOT_EMPTY(empty_regex="de[fg]")("deg")
        self.assertEqual(rtn, ("deg", "Enter a value"))
        rtn = IS_NOT_EMPTY(empty_regex="def")("abc")
        self.assertEqual(rtn, ("abc", None))

    def test_IS_ALPHANUMERIC(self):
        rtn = IS_ALPHANUMERIC()("1")
        self.assertEqual(rtn, ("1", None))
        rtn = IS_ALPHANUMERIC()("")
        self.assertEqual(rtn, ("", None))
        rtn = IS_ALPHANUMERIC()("A_a")
        self.assertEqual(rtn, ("A_a", None))
        rtn = IS_ALPHANUMERIC()("!")
        self.assertEqual(rtn, ("!", "Enter only letters, numbers, and underscore"))

    def test_IS_EMAIL(self):
        rtn = IS_EMAIL()("a@b.com")
        self.assertEqual(rtn, ("a@b.com", None))
        rtn = IS_EMAIL()("abc@def.com")
        self.assertEqual(rtn, ("abc@def.com", None))
        rtn = IS_EMAIL()("abc@3def.com")
        self.assertEqual(rtn, ("abc@3def.com", None))
        rtn = IS_EMAIL()("abc@def.us")
        self.assertEqual(rtn, ("abc@def.us", None))
        rtn = IS_EMAIL()("abc@d_-f.us")
        self.assertEqual(rtn, ("abc@d_-f.us", None))
        rtn = IS_EMAIL()("@def.com")  # missing name
        self.assertEqual(rtn, ("@def.com", "Enter a valid email address"))
        rtn = IS_EMAIL()('"abc@def".com')  # quoted name
        self.assertEqual(rtn, ('"abc@def".com', "Enter a valid email address"))
        rtn = IS_EMAIL()("abc+def.com")  # no @
        self.assertEqual(rtn, ("abc+def.com", "Enter a valid email address"))
        rtn = IS_EMAIL()("abc@def.x")  # one-char TLD
        self.assertEqual(rtn, ("abc@def.x", "Enter a valid email address"))
        rtn = IS_EMAIL()("abc@def.12")  # numeric TLD
        self.assertEqual(rtn, ("abc@def.12", "Enter a valid email address"))
        rtn = IS_EMAIL()("abc@def..com")  # double-dot in domain
        self.assertEqual(rtn, ("abc@def..com", "Enter a valid email address"))
        rtn = IS_EMAIL()("abc@.def.com")  # dot starts domain
        self.assertEqual(rtn, ("abc@.def.com", "Enter a valid email address"))
        rtn = IS_EMAIL()("abc@def.c_m")  # underscore in TLD
        self.assertEqual(rtn, ("abc@def.c_m", "Enter a valid email address"))
        rtn = IS_EMAIL()("NotAnEmail")  # missing @
        self.assertEqual(rtn, ("NotAnEmail", "Enter a valid email address"))
        rtn = IS_EMAIL()("abc@NotAnEmail")  # missing TLD
        self.assertEqual(rtn, ("abc@NotAnEmail", "Enter a valid email address"))
        rtn = IS_EMAIL()("customer/department@example.com")
        self.assertEqual(rtn, ("customer/department@example.com", None))
        rtn = IS_EMAIL()("$A12345@example.com")
        self.assertEqual(rtn, ("$A12345@example.com", None))
        rtn = IS_EMAIL()("!def!xyz%abc@example.com")
        self.assertEqual(rtn, ("!def!xyz%abc@example.com", None))
        rtn = IS_EMAIL()("_Yosemite.Sam@example.com")
        self.assertEqual(rtn, ("_Yosemite.Sam@example.com", None))
        rtn = IS_EMAIL()("~@example.com")
        self.assertEqual(rtn, ("~@example.com", None))
        rtn = IS_EMAIL()(".wooly@example.com")  # dot starts name
        self.assertEqual(rtn, (".wooly@example.com", "Enter a valid email address"))
        rtn = IS_EMAIL()("wo..oly@example.com")  # adjacent dots in name
        self.assertEqual(rtn, ("wo..oly@example.com", "Enter a valid email address"))
        rtn = IS_EMAIL()("pootietang.@example.com")  # dot ends name
        self.assertEqual(
            rtn, ("pootietang.@example.com", "Enter a valid email address")
        )
        rtn = IS_EMAIL()(".@example.com")  # name is bare dot
        self.assertEqual(rtn, (".@example.com", "Enter a valid email address"))
        rtn = IS_EMAIL()("Ima.Fool@example.com")
        self.assertEqual(rtn, ("Ima.Fool@example.com", None))
        rtn = IS_EMAIL()("Ima Fool@example.com")  # space in name
        self.assertEqual(rtn, ("Ima Fool@example.com", "Enter a valid email address"))
        rtn = IS_EMAIL()("localguy@localhost")  # localhost as domain
        self.assertEqual(rtn, ("localguy@localhost", None))
        # test for banned
        rtn = IS_EMAIL(banned=r"^.*\.com(|\..*)$")(
            "localguy@localhost"
        )  # localhost as domain
        self.assertEqual(rtn, ("localguy@localhost", None))
        rtn = IS_EMAIL(banned=r"^.*\.com(|\..*)$")("abc@example.com")
        self.assertEqual(rtn, ("abc@example.com", "Enter a valid email address"))
        # test for forced
        rtn = IS_EMAIL(forced=r"^.*\.edu(|\..*)$")("localguy@localhost")
        self.assertEqual(rtn, ("localguy@localhost", "Enter a valid email address"))
        rtn = IS_EMAIL(forced=r"^.*\.edu(|\..*)$")("localguy@example.edu")
        self.assertEqual(rtn, ("localguy@example.edu", None))
        # test for not a string at all
        rtn = IS_EMAIL(error_message="oops")(42)
        self.assertEqual(rtn, (42, "oops"))

        # test for Internationalized Domain Names, see https://docs.python.org/2/library/codecs.html#module-encodings.idna
        rtn = IS_EMAIL()("web2py@Alliancefrançaise.nu")
        self.assertEqual(rtn, ("web2py@Alliancefrançaise.nu", None))

    def test_IS_LIST_OF_EMAILS(self):
        emails = ["localguy@localhost", "_Yosemite.Sam@example.com"]
        rtn = IS_LIST_OF_EMAILS()(",".join(emails))
        self.assertEqual(rtn, (emails, None))
        rtn = IS_LIST_OF_EMAILS()(";".join(emails))
        self.assertEqual(rtn, (emails, None))
        rtn = IS_LIST_OF_EMAILS()(" ".join(emails))
        self.assertEqual(rtn, (emails, None))
        emails.append("a")
        rtn = IS_LIST_OF_EMAILS()(";".join(emails))
        self.assertEqual(
            rtn, ("localguy@localhost;_Yosemite.Sam@example.com;a", "Invalid emails: a")
        )
        rtn = IS_LIST_OF_EMAILS()("")
        self.assertEqual(rtn, ([], None))
        rtn = IS_LIST_OF_EMAILS().formatter(["test@example.com", "dude@example.com"])
        self.assertEqual(rtn, "test@example.com, dude@example.com")

    def test_IS_URL(self):
        rtn = IS_URL()("http://example.com")
        self.assertEqual(rtn, ("http://example.com", None))
        rtn = IS_URL(error_message="oops")("http://example,com")
        self.assertEqual(rtn, ("http://example,com", "oops"))
        rtn = IS_URL(error_message="oops")(
            "http://www.example.com:8800/a/b/c/d/e/f/g/h"
        )
        self.assertEqual(rtn, ("http://www.example.com:8800/a/b/c/d/e/f/g/h", None))
        rtn = IS_URL(error_message="oops", prepend_scheme="http")("example.com")
        self.assertEqual(rtn, ("http://example.com", None))
        rtn = IS_URL()("http://example.com?q=george&p=22")
        self.assertEqual(rtn, ("http://example.com?q=george&p=22", None))
        rtn = IS_URL(mode="generic", prepend_scheme=None)("example.com")
        self.assertEqual(rtn, ("example.com", None))

    def test_IS_TIME(self):
        rtn = IS_TIME()("21:30")
        self.assertEqual(rtn, (datetime.time(21, 30), None))
        rtn = IS_TIME()("21-30")
        self.assertEqual(rtn, (datetime.time(21, 30), None))
        rtn = IS_TIME()("21.30")
        self.assertEqual(rtn, (datetime.time(21, 30), None))
        rtn = IS_TIME()("21:30:59")
        self.assertEqual(rtn, (datetime.time(21, 30, 59), None))
        rtn = IS_TIME()("5:30")
        self.assertEqual(rtn, (datetime.time(5, 30), None))
        rtn = IS_TIME()("5:30 am")
        self.assertEqual(rtn, (datetime.time(5, 30), None))
        rtn = IS_TIME()("5:30 pm")
        self.assertEqual(rtn, (datetime.time(17, 30), None))
        rtn = IS_TIME()("5:30 whatever")
        self.assertEqual(
            rtn, ("5:30 whatever", "Enter time as hh:mm:ss (seconds, am, pm optional)")
        )
        rtn = IS_TIME()("5:30 20")
        self.assertEqual(
            rtn, ("5:30 20", "Enter time as hh:mm:ss (seconds, am, pm optional)")
        )
        rtn = IS_TIME()("24:30")
        self.assertEqual(
            rtn, ("24:30", "Enter time as hh:mm:ss (seconds, am, pm optional)")
        )
        rtn = IS_TIME()("21:60")
        self.assertEqual(
            rtn, ("21:60", "Enter time as hh:mm:ss (seconds, am, pm optional)")
        )
        rtn = IS_TIME()("21:30::")
        self.assertEqual(
            rtn, ("21:30::", "Enter time as hh:mm:ss (seconds, am, pm optional)")
        )
        rtn = IS_TIME()("")
        self.assertEqual(rtn, ("", "Enter time as hh:mm:ss (seconds, am, pm optional)"))

    def test_IS_DATE(self):
        v = IS_DATE(format="%m/%d/%Y", error_message="oops")
        rtn = v("03/03/2008")
        self.assertEqual(rtn, (datetime.date(2008, 3, 3), None))
        rtn = v("31/03/2008")
        self.assertEqual(rtn, ("31/03/2008", "oops"))
        rtn = IS_DATE(format="%m/%d/%Y", error_message="oops").formatter(
            datetime.date(1834, 12, 14)
        )
        self.assertEqual(rtn, "12/14/1834")

    def test_IS_DATETIME(self):
        v = IS_DATETIME(format="%m/%d/%Y %H:%M", error_message="oops")
        rtn = v("03/03/2008 12:40")
        self.assertEqual(rtn, (datetime.datetime(2008, 3, 3, 12, 40), None))
        rtn = v("31/03/2008 29:40")
        self.assertEqual(rtn, ("31/03/2008 29:40", "oops"))
        # Test timezone is removed and value is properly converted
        #
        # https://github.com/web2py/web2py/issues/1094

        class DummyTimezone(datetime.tzinfo):
            ONE = datetime.timedelta(hours=1)

            def utcoffset(self, dt):
                return DummyTimezone.ONE

            def tzname(self, dt):
                return "UTC+1"

            def dst(self, dt):
                return DummyTimezone.ONE

            def localize(self, dt, is_dst=False):
                return dt.replace(tzinfo=self)

        v = IS_DATETIME(
            format="%Y-%m-%d %H:%M", error_message="oops", timezone=DummyTimezone()
        )
        rtn = v("1982-12-14 08:00")
        self.assertEqual(rtn, (datetime.datetime(1982, 12, 14, 7, 0), None))

    def test_IS_DATE_IN_RANGE(self):
        v = IS_DATE_IN_RANGE(
            minimum=datetime.date(2008, 1, 1),
            maximum=datetime.date(2009, 12, 31),
            format="%m/%d/%Y",
            error_message="oops",
        )

        rtn = v("03/03/2008")
        self.assertEqual(rtn, (datetime.date(2008, 3, 3), None))
        rtn = v("03/03/2010")
        self.assertEqual(rtn, ("03/03/2010", "oops"))
        rtn = v(datetime.date(2008, 3, 3))
        self.assertEqual(rtn, (datetime.date(2008, 3, 3), None))
        rtn = v(datetime.date(2010, 3, 3))
        self.assertEqual(rtn, (datetime.date(2010, 3, 3), "oops"))
        v = IS_DATE_IN_RANGE(maximum=datetime.date(2009, 12, 31), format="%m/%d/%Y")
        rtn = v("03/03/2010")
        self.assertEqual(rtn, ("03/03/2010", "Enter date on or before 12/31/2009"))
        v = IS_DATE_IN_RANGE(minimum=datetime.date(2008, 1, 1), format="%m/%d/%Y")
        rtn = v("03/03/2007")
        self.assertEqual(rtn, ("03/03/2007", "Enter date on or after 01/01/2008"))
        v = IS_DATE_IN_RANGE(
            minimum=datetime.date(2008, 1, 1),
            maximum=datetime.date(2009, 12, 31),
            format="%m/%d/%Y",
        )
        rtn = v("03/03/2007")
        self.assertEqual(
            rtn, ("03/03/2007", "Enter date in range 01/01/2008 12/31/2009")
        )

    def test_IS_DATETIME_IN_RANGE(self):
        v = IS_DATETIME_IN_RANGE(
            minimum=datetime.datetime(2008, 1, 1, 12, 20),
            maximum=datetime.datetime(2009, 12, 31, 12, 20),
            format="%m/%d/%Y %H:%M",
            error_message="oops",
        )
        rtn = v("03/03/2008 12:40")
        self.assertEqual(rtn, (datetime.datetime(2008, 3, 3, 12, 40), None))
        rtn = v("03/03/2010 10:34")
        self.assertEqual(rtn, ("03/03/2010 10:34", "oops"))
        rtn = v(datetime.datetime(2008, 3, 3, 0, 0))
        self.assertEqual(rtn, (datetime.datetime(2008, 3, 3, 0, 0), None))
        rtn = v(datetime.datetime(2010, 3, 3, 0, 0))
        self.assertEqual(rtn, (datetime.datetime(2010, 3, 3, 0, 0), "oops"))
        v = IS_DATETIME_IN_RANGE(
            maximum=datetime.datetime(2009, 12, 31, 12, 20), format="%m/%d/%Y %H:%M:%S"
        )
        rtn = v("03/03/2010 12:20:00")
        self.assertEqual(
            rtn,
            (
                "03/03/2010 12:20:00",
                "Enter date and time on or before 12/31/2009 12:20:00",
            ),
        )
        v = IS_DATETIME_IN_RANGE(
            minimum=datetime.datetime(2008, 1, 1, 12, 20), format="%m/%d/%Y %H:%M:%S"
        )
        rtn = v("03/03/2007 12:20:00")
        self.assertEqual(
            rtn,
            (
                "03/03/2007 12:20:00",
                "Enter date and time on or after 01/01/2008 12:20:00",
            ),
        )
        v = IS_DATETIME_IN_RANGE(
            minimum=datetime.datetime(2008, 1, 1, 12, 20),
            maximum=datetime.datetime(2009, 12, 31, 12, 20),
            format="%m/%d/%Y %H:%M:%S",
        )
        rtn = v("03/03/2007 12:20:00")
        self.assertEqual(
            rtn,
            (
                "03/03/2007 12:20:00",
                "Enter date and time in range 01/01/2008 12:20:00 12/31/2009 12:20:00",
            ),
        )
        v = IS_DATETIME_IN_RANGE(
            maximum=datetime.datetime(2009, 12, 31, 12, 20),
            format="%Y-%m-%d %H:%M:%S",
            error_message="oops",
        )
        rtn = v("clearly not a date")
        self.assertEqual(rtn, ("clearly not a date", "oops"))

    def test_IS_LIST_OF_STRINRGS(self):
        rtn = IS_LIST_OF_STRINGS()("abc")
        self.assertEqual(rtn, (["abc"], None))
        rtn = IS_LIST_OF_STRINGS()(["a", "b", "c"])
        self.assertEqual(rtn, (["a", "b", "c"], None))
        rtn = IS_LIST_OF_STRINGS()("a,b,c")
        self.assertEqual(rtn, (["a", "b", "c"], None))
        rtn = IS_LIST_OF_STRINGS()("a,b;c")
        self.assertEqual(rtn, (["a", "b", "c"], None))
        rtn = IS_LIST_OF_STRINGS()('a,"hello world"; c')
        self.assertEqual(rtn, (["a", "hello world", "c"], None))
        rtn = IS_LIST_OF_STRINGS()('a,"hello\\" world"; c')
        self.assertEqual(rtn, (["a", 'hello" world', "c"], None))
        rtn = IS_LIST_OF_STRINGS().formatter(["a", 'hello" world', "c"])
        self.assertEqual(rtn, 'a, "hello\\" world", c')

    def test_IS_LIST_OF_INTS(self):
        # can validate a single number
        rtn = IS_LIST_OF_INTS()(1)
        self.assertEqual(rtn, ([1], None))
        # can validate lists
        rtn = IS_LIST_OF_INTS()([1])
        self.assertEqual(rtn, ([1], None))
        rtn = IS_LIST_OF_INTS()([1, 2, 3])
        self.assertEqual(rtn, ([1, 2, 3], None))
        # can validate numbers as strings
        rtn = IS_LIST_OF_INTS()("1")
        self.assertEqual(rtn, ([1], None))
        # can validate json lists
        rtn = IS_LIST_OF_INTS()("[1, 2, 3]")
        self.assertEqual(rtn, ([1, 2, 3], None))
        rtn = IS_LIST_OF_INTS()('["1", "2", "3"]')
        self.assertEqual(rtn, ([1, 2, 3], None))
        # can validate comma separated of semicolon separated numbers
        rtn = IS_LIST_OF_INTS()("1, 2, 3")
        self.assertEqual(rtn, ([1, 2, 3], None))
        rtn = IS_LIST_OF_INTS()("1, 2; 3")
        self.assertEqual(rtn, ([1, 2, 3], None))
        # can validate comma separted quoted numbers
        rtn = IS_LIST_OF_INTS()('1, "2", 3')
        self.assertEqual(rtn, ([1, 2, 3], None))
        # check invalid values
        rtn = IS_LIST_OF_INTS()("abc")
        self.assertEqual(rtn, ("abc", "Enter a list of integers"))
        rtn = IS_LIST_OF_INTS()("1, abc, 3")
        self.assertEqual(rtn, ("1, abc, 3", "Enter a list of integers"))
        rtn = IS_LIST_OF_INTS()([1, "abc", 3])
        self.assertEqual(rtn, ([1, "abc", 3], "Enter a list of integers"))
        # check formatting
        rtn = IS_LIST_OF_INTS().formatter([1, 2, 3])
        self.assertEqual(rtn, "1, 2, 3")

    def test_IS_LIST_OF(self):
        values = [0, 1, 2, 3, 4]
        rtn = IS_LIST_OF(IS_INT_IN_RANGE(0, 10))(values)
        self.assertEqual(rtn, (values, None))
        values.append(11)
        rtn = IS_LIST_OF(IS_INT_IN_RANGE(0, 10))(values)
        self.assertEqual(rtn, (values, "Enter an integer between 0 and 9"))
        rtn = IS_LIST_OF(IS_INT_IN_RANGE(0, 10))(1)
        self.assertEqual(rtn, ([1], None))
        rtn = IS_LIST_OF(IS_INT_IN_RANGE(0, 10), minimum=10)([1, 2])
        self.assertEqual(rtn, ([1, 2], "Minimum length is 10"))
        rtn = IS_LIST_OF(IS_INT_IN_RANGE(0, 10), maximum=2)([1, 2, 3])
        self.assertEqual(rtn, ([1, 2, 3], "Maximum length is 2"))
        # regression test for issue 742
        rtn = IS_LIST_OF(minimum=1)("")
        self.assertEqual(rtn, ("", "Minimum length is 1"))

    def test_IS_LOWER(self):
        rtn = IS_LOWER()("ABC")
        self.assertEqual(rtn, ("abc", None))
        rtn = IS_LOWER()(b"ABC")
        self.assertEqual(rtn, (b"abc", None))
        rtn = IS_LOWER()("Ñ")
        self.assertEqual(rtn, ("ñ", None))

    def test_IS_UPPER(self):
        rtn = IS_UPPER()("abc")
        self.assertEqual(rtn, ("ABC", None))
        rtn = IS_UPPER()(b"abc")
        self.assertEqual(rtn, (b"ABC", None))
        rtn = IS_UPPER()("ñ")
        self.assertEqual(rtn, ("Ñ", None))

    def test_IS_SLUG(self):
        rtn = IS_SLUG()("abc123")
        self.assertEqual(rtn, ("abc123", None))
        rtn = IS_SLUG()("ABC123")
        self.assertEqual(rtn, ("abc123", None))
        rtn = IS_SLUG()("abc-123")
        self.assertEqual(rtn, ("abc-123", None))
        rtn = IS_SLUG()("abc--123")
        self.assertEqual(rtn, ("abc-123", None))
        rtn = IS_SLUG()("abc 123")
        self.assertEqual(rtn, ("abc-123", None))
        rtn = IS_SLUG()("abc\t_123")
        self.assertEqual(rtn, ("abc-123", None))
        rtn = IS_SLUG()("-abc-")
        self.assertEqual(rtn, ("abc", None))
        rtn = IS_SLUG()("--a--b--_ -c--")
        self.assertEqual(rtn, ("a-b-c", None))
        rtn = IS_SLUG()("abc&amp;123")
        self.assertEqual(rtn, ("abc123", None))
        rtn = IS_SLUG()("abc&amp;123&amp;def")
        self.assertEqual(rtn, ("abc123def", None))
        rtn = IS_SLUG()("ñ")
        self.assertEqual(rtn, ("n", None))
        rtn = IS_SLUG(maxlen=4)("abc123")
        self.assertEqual(rtn, ("abc1", None))
        rtn = IS_SLUG()("abc_123")
        self.assertEqual(rtn, ("abc-123", None))
        rtn = IS_SLUG(keep_underscores=False)("abc_123")
        self.assertEqual(rtn, ("abc-123", None))
        rtn = IS_SLUG(keep_underscores=True)("abc_123")
        self.assertEqual(rtn, ("abc_123", None))
        rtn = IS_SLUG(check=False)("abc")
        self.assertEqual(rtn, ("abc", None))
        rtn = IS_SLUG(check=True)("abc")
        self.assertEqual(rtn, ("abc", None))
        rtn = IS_SLUG(check=False)("a bc")
        self.assertEqual(rtn, ("a-bc", None))
        rtn = IS_SLUG(check=True)("a bc")
        self.assertEqual(rtn, ("a bc", "Must be slug"))

    def test_ANY_OF(self):
        rtn = ANY_OF([IS_EMAIL(), IS_ALPHANUMERIC()])("a@b.co")
        self.assertEqual(rtn, ("a@b.co", None))
        rtn = ANY_OF([IS_EMAIL(), IS_ALPHANUMERIC()])("abco")
        self.assertEqual(rtn, ("abco", None))
        rtn = ANY_OF([IS_EMAIL(), IS_ALPHANUMERIC()])("@ab.co")
        self.assertEqual(rtn, ("@ab.co", "Enter only letters, numbers, and underscore"))
        rtn = ANY_OF([IS_ALPHANUMERIC(), IS_EMAIL()])("@ab.co")
        self.assertEqual(rtn, ("@ab.co", "Enter a valid email address"))
        rtn = ANY_OF([IS_DATE(), IS_EMAIL()])("a@b.co")
        self.assertEqual(rtn, ("a@b.co", None))
        rtn = ANY_OF([IS_DATE(), IS_EMAIL()])("1982-12-14")
        self.assertEqual(rtn, (datetime.date(1982, 12, 14), None))
        rtn = ANY_OF([IS_DATE(format="%m/%d/%Y"), IS_EMAIL()]).formatter(
            datetime.date(1834, 12, 14)
        )
        self.assertEqual(rtn, "12/14/1834")

    class CustomValidator(Validator):
        """Custom Validator not overloading .validate()"""

        def __call__(self, value):
            return (value, None) if value else (value, "invalid!")

    @staticmethod
    def custom_validator_func(value):
        """Validation function instead of Validator subclass"""
        return (value, None) if value else (value, "invalid!")

    def test_IS_EMPTY_OR(self):
        rtn = IS_EMPTY_OR(IS_EMAIL())("abc@def.com")
        self.assertEqual(rtn, ("abc@def.com", None))
        rtn = IS_EMPTY_OR(IS_EMAIL())("   ")
        self.assertEqual(rtn, (None, None))
        rtn = IS_EMPTY_OR(IS_EMAIL(), null="abc")("   ")
        self.assertEqual(rtn, ("abc", None))
        rtn = IS_EMPTY_OR(IS_EMAIL(), null="abc", empty_regex="def")("def")
        self.assertEqual(rtn, ("abc", None))
        rtn = IS_EMPTY_OR(IS_EMAIL())("abc")
        self.assertEqual(rtn, ("abc", "Enter a valid email address"))
        rtn = IS_EMPTY_OR(IS_EMAIL())(" abc ")
        self.assertEqual(rtn, (" abc ", "Enter a valid email address"))
        rtn = IS_EMPTY_OR(
            IS_IN_SET([("id1", "first label"), ("id2", "second label")], zero="zero")
        ).options(zero=False)
        self.assertEqual(
            rtn, [("", ""), ("id1", "first label"), ("id2", "second label")]
        )
        rtn = IS_EMPTY_OR(
            IS_IN_SET([("id1", "first label"), ("id2", "second label")], zero="zero")
        ).options()
        self.assertEqual(
            rtn, [("", "zero"), ("id1", "first label"), ("id2", "second label")]
        )
        rtn = IS_EMPTY_OR((IS_LOWER(), IS_EMAIL()))("AAA")
        self.assertEqual(rtn, ("AAA", "Enter a valid email address"))
        rtn = IS_EMPTY_OR([IS_LOWER(), IS_EMAIL()])("AAA")
        self.assertEqual(rtn, ("AAA", "Enter a valid email address"))
        rtn = IS_EMPTY_OR(self.CustomValidator())(1)
        self.assertEqual(rtn, (1, None))
        rtn = IS_EMPTY_OR(self.CustomValidator())(0)
        self.assertEqual(rtn, (0, "invalid!"))
        rtn = IS_EMPTY_OR(self.custom_validator_func)(1)
        self.assertEqual(rtn, (1, None))
        rtn = IS_EMPTY_OR(self.custom_validator_func)(0)
        self.assertEqual(rtn, (0, "invalid!"))

    def test_CLEANUP(self):
        rtn = CLEANUP()("helloò")
        self.assertEqual(rtn, ("hello", None))

    def test_CRYPT(self):
        rtn = str(CRYPT(digest_alg="md5", salt=True)("test")[0])
        self.myassertRegex(rtn, r"^md5\$.{16}\$.{32}$")
        rtn = str(CRYPT(digest_alg="sha1", salt=True)("test")[0])
        self.myassertRegex(rtn, r"^sha1\$.{16}\$.{40}$")
        rtn = str(CRYPT(digest_alg="sha256", salt=True)("test")[0])
        self.myassertRegex(rtn, r"^sha256\$.{16}\$.{64}$")
        rtn = str(CRYPT(digest_alg="sha384", salt=True)("test")[0])
        self.myassertRegex(rtn, r"^sha384\$.{16}\$.{96}$")
        rtn = str(CRYPT(digest_alg="sha512", salt=True)("test")[0])
        self.myassertRegex(rtn, r"^sha512\$.{16}\$.{128}$")
        alg = "pbkdf2(1000,20,sha512)"
        rtn = str(CRYPT(digest_alg=alg, salt=True)("test")[0])
        self.myassertRegex(rtn, r"^pbkdf2\(1000,20,sha512\)\$.{16}\$.{40}$")
        rtn = str(CRYPT(digest_alg="md5", key="mykey", salt=True)("test")[0])
        self.myassertRegex(rtn, r"^md5\$.{16}\$.{32}$")
        a = str(CRYPT(digest_alg="sha1", salt=False)("test")[0])
        self.assertEqual(CRYPT(digest_alg="sha1", salt=False)("test")[0], a)
        self.assertEqual(CRYPT(digest_alg="sha1", salt=False)("test")[0], a[6:])
        self.assertEqual(CRYPT(digest_alg="md5", salt=False)("test")[0], a)
        self.assertEqual(CRYPT(digest_alg="md5", salt=False)("test")[0], a[6:])

    def test_IS_STRONG(self):
        rtn = IS_STRONG(es=True)("Abcd1234")
        self.assertEqual(
            rtn,
            (
                "Abcd1234",
                "Must include at least 1 of the following: ~!@#$%^&*()_+-=?<>,.:;{}[]|",
            ),
        )
        rtn = IS_STRONG(es=True)("Abcd1234!")
        self.assertEqual(rtn, ("Abcd1234!", None))
        rtn = IS_STRONG(es=True, entropy=1)("a")
        self.assertEqual(rtn, ("a", None))
        rtn = IS_STRONG(es=True, entropy=1, min=2)("a")
        self.assertEqual(rtn, ("a", "Minimum length is 2"))
        rtn = IS_STRONG(es=True, entropy=100)("abc123")
        self.assertEqual(rtn, ("abc123", "Password too simple (32.35/100)"))
        rtn = IS_STRONG(es=True, entropy=100)("and")
        self.assertEqual(rtn, ("and", "Password too simple (14.57/100)"))
        rtn = IS_STRONG(es=True, entropy=100)("aaa")
        self.assertEqual(rtn, ("aaa", "Password too simple (14.42/100)"))
        rtn = IS_STRONG(es=True, entropy=100)("a1d")
        self.assertEqual(rtn, ("a1d", "Password too simple (15.97/100)"))
        rtn = IS_STRONG(es=True, entropy=100)("añd")
        if PY2:
            self.assertEqual(rtn, ("a\xc3\xb1d", "Password too simple (32.66/100)"))
        else:
            self.assertEqual(rtn, ("añd", "Password too simple (31.26/100)"))
        rtn = IS_STRONG()("********")
        self.assertEqual(rtn, ("********", None))
        rtn = IS_STRONG(es=True, max=4)("abcde")
        self.assertEqual(
            rtn,
            (
                "abcde",
                "|".join(
                    [
                        "Minimum length is 8",
                        "Maximum length is 4",
                        "Must include at least 1 of the following: ~!@#$%^&*()_+-=?<>,.:;{}[]|",
                        "Must include at least 1 uppercase",
                        "Must include at least 1 number",
                    ]
                ),
            ),
        )
        rtn = IS_STRONG(es=True)("abcde")
        self.assertEqual(
            rtn,
            (
                "abcde",
                "|".join(
                    [
                        "Minimum length is 8",
                        "Must include at least 1 of the following: ~!@#$%^&*()_+-=?<>,.:;{}[]|",
                        "Must include at least 1 uppercase",
                        "Must include at least 1 number",
                    ]
                ),
            ),
        )
        rtn = IS_STRONG(upper=0, lower=0, number=0, es=True)("Abcde1")
        self.assertEqual(
            rtn,
            (
                "Abcde1",
                "|".join(
                    [
                        "Minimum length is 8",
                        "Must include at least 1 of the following: ~!@#$%^&*()_+-=?<>,.:;{}[]|",
                        "May not include any uppercase letters",
                        "May not include any lowercase letters",
                        "May not include any numbers",
                    ]
                ),
            ),
        )
        rtn = IS_STRONG(special=0, es=True)("Abcde1!")
        self.assertEqual(
            rtn,
            (
                "Abcde1!",
                "|".join(
                    [
                        "Minimum length is 8",
                        "May not contain any of the following: ~!@#$%^&*()_+-=?<>,.:;{}[]|",
                    ]
                ),
            ),
        )
        rtn = IS_STRONG(upper=False, number=False, special=False, es=True)("Abcde1!")
        self.assertEqual(rtn, ("Abcde1!", "Minimum length is 8"))

    def test_IS_IMAGE(self):
        class DummyImageFile(object):
            def __init__(self, filename, ext, width, height):
                import struct
                from io import BytesIO

                self.filename = filename + "." + ext
                self.file = BytesIO()
                if ext == "bmp":
                    self.file.write(b"BM")
                    self.file.write(b" " * 16)
                    self.file.write(struct.pack("<LL", width, height))
                elif ext == "gif":
                    self.file.write(b"GIF87a")
                    self.file.write(struct.pack("<HHB", width, height, 0))
                elif ext in ("jpg", "jpeg"):
                    self.file.write(b"\xff\xd8")
                    self.file.write(struct.pack("!BBH", 0xFF, 0xC0, 5))
                    self.file.write(struct.pack("!xHH", height, width))
                elif ext == "png":
                    self.file.write(b"\211PNG\r\n\032\n")
                    self.file.write(b" " * 4)
                    self.file.write(b"IHDR")
                    self.file.write(struct.pack("!LL", width, height))
                self.file.seek(0)

        img = DummyImageFile("test", "bmp", 50, 100)
        rtn = IS_IMAGE()(img)
        self.assertEqual(rtn, (img, None))
        rtn = IS_IMAGE(error_message="oops", maxsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", minsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", aspectratio=(1, 1))(img)
        self.assertEqual(rtn, (img, "oops"))

        img = DummyImageFile("test", "gif", 50, 100)
        rtn = IS_IMAGE()(img)
        self.assertEqual(rtn, (img, None))
        rtn = IS_IMAGE(error_message="oops", maxsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", minsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", aspectratio=(1, 1))(img)
        self.assertEqual(rtn, (img, "oops"))

        img = DummyImageFile("test", "jpeg", 50, 100)
        rtn = IS_IMAGE()(img)
        self.assertEqual(rtn, (img, None))
        rtn = IS_IMAGE(error_message="oops", maxsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", minsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", aspectratio=(1, 1))(img)
        self.assertEqual(rtn, (img, "oops"))

        img = DummyImageFile("test", "png", 50, 100)
        rtn = IS_IMAGE()(img)
        self.assertEqual(rtn, (img, None))
        rtn = IS_IMAGE(error_message="oops", maxsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", minsize=(100, 50))(img)
        self.assertEqual(rtn, (img, "oops"))
        rtn = IS_IMAGE(error_message="oops", aspectratio=(1, 1))(img)
        self.assertEqual(rtn, (img, "oops"))

        img = DummyImageFile("test", "xls", 50, 100)
        rtn = IS_IMAGE(error_message="oops")(img)
        self.assertEqual(rtn, (img, "oops"))

    def test_IS_FILE(self):
        from io import BytesIO

        def gen_fake(filename):
            formdata_file_data = (
                """
---123
Content-Disposition: form-data; name="key2"

value2y
---123
Content-Disposition: form-data; name="file_attach"; filename="%s"
Content-Type: text/plain

this is the content of the fake file

---123--
"""
                % filename
            )
            formdata_file_environ = {
                "CONTENT_LENGTH": str(len(formdata_file_data)),
                "CONTENT_TYPE": "multipart/form-data; boundary=-123",
                "QUERY_STRING": "key1=value1&key2=value2x",
                "REQUEST_METHOD": "POST",
            }
            return cgi.FieldStorage(
                fp=BytesIO(to_bytes(formdata_file_data)), environ=formdata_file_environ
            )["file_attach"]

        fake = gen_fake("example.pdf")
        rtn = IS_FILE(extension="pdf")(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("example.gif")
        rtn = IS_FILE(extension="pdf")(fake)
        self.assertEqual(rtn, (fake, "Enter valid filename"))
        fake = gen_fake("multiple.pdf")
        rtn = IS_FILE(extension=["pdf", "png"])(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("multiple.png")
        rtn = IS_FILE(extension=["pdf", "png"])(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("multiple.gif")
        rtn = IS_FILE(extension=["pdf", "png"])(fake)
        self.assertEqual(rtn, (fake, "Enter valid filename"))
        fake = gen_fake("backup2014.tar.gz")
        rtn = IS_FILE(
            filename=re.compile("backup.*"), extension="tar.gz", lastdot=False
        )(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("README")
        rtn = IS_FILE(filename="README", extension="", case=0)(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("readme")
        rtn = IS_FILE(filename="README", extension="", case=0)(fake)
        self.assertEqual(rtn, (fake, "Enter valid filename"))
        fake = gen_fake("readme")
        rtn = IS_FILE(filename="README", case=2)(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("README")
        rtn = IS_FILE(filename="README", case=2)(fake)
        self.assertEqual(rtn, (fake, None))
        rtn = IS_FILE(extension="pdf")("example.pdf")
        self.assertEqual(rtn, ("example.pdf", "Enter valid filename"))

    def test_IS_UPLOAD_FILENAME(self):
        from io import BytesIO

        def gen_fake(filename):
            formdata_file_data = (
                """
---123
Content-Disposition: form-data; name="key2"

value2y
---123
Content-Disposition: form-data; name="file_attach"; filename="%s"
Content-Type: text/plain

this is the content of the fake file

---123--
"""
                % filename
            )
            formdata_file_environ = {
                "CONTENT_LENGTH": str(len(formdata_file_data)),
                "CONTENT_TYPE": "multipart/form-data; boundary=-123",
                "QUERY_STRING": "key1=value1&key2=value2x",
                "REQUEST_METHOD": "POST",
            }
            return cgi.FieldStorage(
                fp=BytesIO(to_bytes(formdata_file_data)), environ=formdata_file_environ
            )["file_attach"]

        fake = gen_fake("example.pdf")
        rtn = IS_UPLOAD_FILENAME(extension="pdf")(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("example.gif")
        rtn = IS_UPLOAD_FILENAME(extension="pdf")(fake)
        self.assertEqual(rtn, (fake, "Enter valid filename"))
        fake = gen_fake("backup2014.tar.gz")
        rtn = IS_UPLOAD_FILENAME(
            filename="backup.*", extension="tar.gz", lastdot=False
        )(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("README")
        rtn = IS_UPLOAD_FILENAME(filename="^README$", extension="^$", case=0)(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("readme")
        rtn = IS_UPLOAD_FILENAME(filename="^README$", extension="^$", case=0)(fake)
        self.assertEqual(rtn, (fake, "Enter valid filename"))
        fake = gen_fake("readme")
        rtn = IS_UPLOAD_FILENAME(filename="README", case=2)(fake)
        self.assertEqual(rtn, (fake, None))
        fake = gen_fake("README")
        rtn = IS_UPLOAD_FILENAME(filename="README", case=2)(fake)
        self.assertEqual(rtn, (fake, None))
        rtn = IS_UPLOAD_FILENAME(extension="pdf")("example.pdf")
        self.assertEqual(rtn, ("example.pdf", "Enter valid filename"))

    def test_IS_IPV4(self):
        rtn = IS_IPV4()("1.2.3.4")
        self.assertEqual(rtn, ("1.2.3.4", None))
        rtn = IS_IPV4()("255.255.255.255")
        self.assertEqual(rtn, ("255.255.255.255", None))
        rtn = IS_IPV4()("1.2.3.4 ")
        self.assertEqual(rtn, ("1.2.3.4 ", "Enter valid IPv4 address"))
        rtn = IS_IPV4()("1.2.3.4.5")
        self.assertEqual(rtn, ("1.2.3.4.5", "Enter valid IPv4 address"))
        rtn = IS_IPV4()("123.123")
        self.assertEqual(rtn, ("123.123", "Enter valid IPv4 address"))
        rtn = IS_IPV4()("1111.2.3.4")
        self.assertEqual(rtn, ("1111.2.3.4", "Enter valid IPv4 address"))
        rtn = IS_IPV4()("0111.2.3.4")
        self.assertEqual(rtn, ("0111.2.3.4", "Enter valid IPv4 address"))
        rtn = IS_IPV4()("256.2.3.4")
        self.assertEqual(rtn, ("256.2.3.4", "Enter valid IPv4 address"))
        rtn = IS_IPV4()("300.2.3.4")
        self.assertEqual(rtn, ("300.2.3.4", "Enter valid IPv4 address"))
        rtn = IS_IPV4(minip="1.2.3.4", maxip="1.2.3.4")("1.2.3.4")
        self.assertEqual(rtn, ("1.2.3.4", None))
        rtn = IS_IPV4(minip="1.2.3.5", maxip="1.2.3.9", error_message="bad ip")(
            "1.2.3.4"
        )
        self.assertEqual(rtn, ("1.2.3.4", "bad ip"))
        rtn = IS_IPV4(maxip="1.2.3.4", invert=True)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", None))
        rtn = IS_IPV4(maxip="1.2.3.4", invert=True)("1.2.3.4")
        self.assertEqual(rtn, ("1.2.3.4", "Enter valid IPv4 address"))
        rtn = IS_IPV4(is_localhost=True)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", None))
        rtn = IS_IPV4(is_localhost=True)("1.2.3.4")
        self.assertEqual(rtn, ("1.2.3.4", "Enter valid IPv4 address"))
        rtn = IS_IPV4(is_localhost=False)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", "Enter valid IPv4 address"))
        rtn = IS_IPV4(maxip="100.0.0.0", is_localhost=True)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", "Enter valid IPv4 address"))

    def test_IS_IPV6(self):
        rtn = IS_IPV6()("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", None))
        rtn = IS_IPV6()("192.168.1.1")
        self.assertEqual(rtn, ("192.168.1.1", "Enter valid IPv6 address"))
        rtn = IS_IPV6(error_message="bad ip")("192.168.1.1")
        self.assertEqual(rtn, ("192.168.1.1", "bad ip"))
        rtn = IS_IPV6(is_link_local=True)("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", None))
        rtn = IS_IPV6(is_link_local=False)("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", "Enter valid IPv6 address"))
        rtn = IS_IPV6(is_link_local=True)("2001::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::126c:8ffa:fe22:b3af", "Enter valid IPv6 address"))
        rtn = IS_IPV6(is_multicast=True)("2001::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::126c:8ffa:fe22:b3af", "Enter valid IPv6 address"))
        rtn = IS_IPV6(is_multicast=True)("ff00::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("ff00::126c:8ffa:fe22:b3af", None))
        # with py3.ipaddress '2001::126c:8ffa:fe22:b3af' is considered private
        # with py2.ipaddress '2001::126c:8ffa:fe22:b3af' is considered private
        # with gluon.contrib.ipaddr(both current and trunk) is not considered private
        rtn = IS_IPV6(is_routeable=False)("2001::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::126c:8ffa:fe22:b3af", None))
        rtn = IS_IPV6(is_routeable=True)("ff00::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("ff00::126c:8ffa:fe22:b3af", "Enter valid IPv6 address"))
        rtn = IS_IPV6(subnets="2001::/32")("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", None))
        rtn = IS_IPV6(subnets="fb00::/8")("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", "Enter valid IPv6 address"))
        rtn = IS_IPV6(subnets=["fc00::/8", "2001::/32"])("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", None))
        rtn = IS_IPV6(subnets="invalidsubnet")("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", "invalid subnet provided"))

    def test_IS_IPADDRESS(self):
        rtn = IS_IPADDRESS()("192.168.1.5")
        self.assertEqual(rtn, ("192.168.1.5", None))
        rtn = IS_IPADDRESS(is_ipv6=False)("192.168.1.5")
        self.assertEqual(rtn, ("192.168.1.5", None))
        rtn = IS_IPADDRESS()("255.255.255.255")
        self.assertEqual(rtn, ("255.255.255.255", None))
        rtn = IS_IPADDRESS()("192.168.1.5 ")
        self.assertEqual(rtn, ("192.168.1.5 ", "Enter valid IP address"))
        rtn = IS_IPADDRESS()("192.168.1.1.5")
        self.assertEqual(rtn, ("192.168.1.1.5", "Enter valid IP address"))
        rtn = IS_IPADDRESS()("123.123")
        self.assertEqual(rtn, ("123.123", "Enter valid IP address"))
        rtn = IS_IPADDRESS()("1111.2.3.4")
        self.assertEqual(rtn, ("1111.2.3.4", "Enter valid IP address"))
        rtn = IS_IPADDRESS()("0111.2.3.4")
        self.assertEqual(rtn, ("0111.2.3.4", "Enter valid IP address"))
        rtn = IS_IPADDRESS()("256.2.3.4")
        self.assertEqual(rtn, ("256.2.3.4", "Enter valid IP address"))
        rtn = IS_IPADDRESS()("300.2.3.4")
        self.assertEqual(rtn, ("300.2.3.4", "Enter valid IP address"))
        rtn = IS_IPADDRESS(minip="192.168.1.0", maxip="192.168.1.255")("192.168.1.100")
        self.assertEqual(rtn, ("192.168.1.100", None))
        rtn = IS_IPADDRESS(minip="1.2.3.5", maxip="1.2.3.9", error_message="bad ip")(
            "1.2.3.4"
        )
        self.assertEqual(rtn, ("1.2.3.4", "bad ip"))
        rtn = IS_IPADDRESS(maxip="1.2.3.4", invert=True)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", None))
        rtn = IS_IPADDRESS(maxip="192.168.1.4", invert=True)("192.168.1.4")
        self.assertEqual(rtn, ("192.168.1.4", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_localhost=True)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", None))
        rtn = IS_IPADDRESS(is_localhost=True)("192.168.1.10")
        self.assertEqual(rtn, ("192.168.1.10", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_localhost=False)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", "Enter valid IP address"))
        rtn = IS_IPADDRESS(maxip="100.0.0.0", is_localhost=True)("127.0.0.1")
        self.assertEqual(rtn, ("127.0.0.1", "Enter valid IP address"))
        rtn = IS_IPADDRESS()("aaa")
        self.assertEqual(rtn, ("aaa", "Enter valid IP address"))

        rtn = IS_IPADDRESS()("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", None))
        rtn = IS_IPADDRESS(is_ipv4=False)("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", None))
        rtn = IS_IPADDRESS()("fe80::126c:8ffa:fe22:b3af  ")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af  ", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_ipv4=True)("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_ipv6=True)("192.168.1.1")
        self.assertEqual(rtn, ("192.168.1.1", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_ipv6=True, error_message="bad ip")("192.168.1.1")
        self.assertEqual(rtn, ("192.168.1.1", "bad ip"))
        rtn = IS_IPADDRESS(is_link_local=True)("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", None))
        rtn = IS_IPADDRESS(is_link_local=False)("fe80::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("fe80::126c:8ffa:fe22:b3af", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_link_local=True)("2001::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::126c:8ffa:fe22:b3af", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_multicast=True)("2001::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::126c:8ffa:fe22:b3af", "Enter valid IP address"))
        rtn = IS_IPADDRESS(is_multicast=True)("ff00::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("ff00::126c:8ffa:fe22:b3af", None))
        rtn = IS_IPADDRESS(is_routeable=True)("ff00::126c:8ffa:fe22:b3af")
        self.assertEqual(rtn, ("ff00::126c:8ffa:fe22:b3af", "Enter valid IP address"))
        rtn = IS_IPADDRESS(subnets="2001::/32")("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", None))
        rtn = IS_IPADDRESS(subnets="fb00::/8")("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", "Enter valid IP address"))
        rtn = IS_IPADDRESS(subnets=["fc00::/8", "2001::/32"])("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", None))
        rtn = IS_IPADDRESS(subnets="invalidsubnet")("2001::8ffa:fe22:b3af")
        self.assertEqual(rtn, ("2001::8ffa:fe22:b3af", "invalid subnet provided"))

    def test_IS_SAFE(self):
        rtn = IS_SAFE()("<div>test</div>")
        ("<div></div>", None)

        rtn = IS_SAFE()("<div><script>xxx</script></div>")
        ("<div><script>xxx</script></div>", "Unsafe Content")

        rtn = IS_SAFE(mode="sanitize")("<div><script>xxx</script></div>")
        ("<div>xxx</div>", "Unsafe Content")

    def test_IS_LIST_OF(self):
        validator = IS_LIST_OF(IS_EMPTY_OR(IS_DATE()))
        self.assertEqual(validator.validate(""), [])
        self.assertEqual(validator.validate([]), [])
        self.assertEqual(validator.validate([None]), [None])
        self.assertEqual(
            validator.validate(["2025-10-12"]), [datetime.date(2025, 10, 12)]
        )

        # try json
        self.assertEqual(validator.validate("[]"), [])
        self.assertEqual(validator.validate("[null]"), [None])
        self.assertEqual(
            validator.validate('["2025-10-12"]'), [datetime.date(2025, 10, 12)]
        )



from .base import DAL
from .objects import Field
from .validators import (
    IS_INT_IN_RANGE,
    IS_FLOAT_IN_RANGE,
    IS_TIME,
    IS_DATE,
    IS_DATETIME,
    IS_JSON,
)
import re


class QueryParseError(RuntimeError):
    pass


def validate(field, value):
    """Validate the value for the field"""
    error = None
    if (
        field.type == "id"
        or field.type.startswith("reference")
        or field.type.startswith("list:reference")
    ):
        error = IS_INT_IN_RANGE(0)(value)[1]
    elif field.type == "integer" or field.type == "list:integer":
        error = IS_INT_IN_RANGE()(value)[1]
    elif field.type == "double" or field.type.startswith("decimal"):
        error = IS_FLOAT_IN_RANGE()(value)[1]
    elif field.type == "time":
        error = IS_TIME()(value)[1]
    elif field.type == "date":
        error = IS_DATE()(value)[1]
    elif field.type == "datetime":
        error = IS_DATETIME()(value)[1]
    if error:
        raise QueryParseError(f"{value}: {error}")


class QueryBuilder:
    tokens_not = {
        "not",
    }
    tokens_ops = {
        "upper",
        "lower",
        "is null",
        "is not null",
        "is true",
        "is false",
        "==",
        "!=",
        "<",
        ">",
        "<=",
        ">=",
        "belongs",
        "contains",
        "startswith",
    }
    tokens_and_or = {
        "and",
        "or",
    }
    default_token_aliases = {
        "is": "==",
        "is equal": "==",
        "is equal to": "==",
        "is not equal": "!=",
        "is not equal to": "!=",
        "is less": "<",
        "is less than": "<",
        "is greater": ">",
        "is greater than": ">",
        "is less or equal": "<=",
        "is less or equal than": "<=",
        "is equal or less": "<=",
        "is equal or less than": "<=",
        "is greater or equal": ">=",
        "is greater or equal than": ">=",
        "is equal or greater": ">=",
        "is equal or greater than": ">=",
        "is in": "belongs",
        "in": "belongs",
        "starts with": "startswith",
    }
    # regex matching field names, and
    re_token = re.compile(r"^(\w+)\s*(.*)$")
    # regex matching a value or quoted value
    re_value = re.compile(r'^((?:")(?:[^"]*[\\]["])*[^"]*(?:")|[^" ,]*)\s*(.*)$')
    # regex matching repeated spaces
    re_spaces = re.compile(r"\s+")

    def __init__(
        self,
        table,
        field_aliases=None,
        token_aliases=None,
        debug=False,
    ):
        """
        Creates a QueryBuilder object
        params:
        - table: the table object to be searched
        - field_aliases: an optional mapping between desired field names and actual field names.
                         If present only listed fields will be searchable. If only only readable fields.
        - token_aliases: a mapping between expressions like "is equal to" into operations like "==".
        """
        self.table = table
        # we either override all fields or none
        if field_aliases:
            self.fields = {k: table[v] for k, v in field_aliases.items()}
        else:
            self.fields = {f.name.lower(): f for f in table if f.readable}
        # use default token aliases if none provided
        if token_aliases is None:
            token_aliases = QueryBuilder.default_token_aliases
        # build a complete list of tokens insluding aliases
        self.tokens_not = self._augment(token_aliases, QueryBuilder.tokens_not)
        self.tokens_and_or = self._augment(token_aliases, QueryBuilder.tokens_and_or)
        self.tokens_ops = self._augment(token_aliases, QueryBuilder.tokens_ops)
        # build the regexes that depend on tokens
        self.re_not = re.compile(r"^(" + "|".join(self.tokens_not) + r")(\W.*)$")
        self.re_op = re.compile(
            "^("
            + "|".join(
                t.replace(" ", r"\s+") for t in sorted(self.tokens_ops, reverse=True)
            )
            + r")\s*(.*)$"
        )
        # true or false
        self.debug = debug

    @staticmethod
    def _augment(aliases, original):
        """returns a dict of k:v for k,v in aliases and v in original"""
        output = {k: k for k in original}
        if aliases:
            output.update({k: v for k, v in aliases.items() if v in original})
        return output

    @staticmethod
    def _find_closing_bracket(text):
        """Finds the end of a bracketed expression"""
        if text[:1] != "(":
            raise QueryParseError("Internal error: missing start bracket")
        level = 0
        quoted = False
        prev_c = None
        for i, c in enumerate(text):
            if not quoted:
                if c == '"':
                    quoted = True
                elif c == "(":
                    level += 1
                elif c == ")":
                    level -= 1
                    if level == 0:
                        return i
            elif c == '"':
                if prev_c != "\\":
                    quoted = False
            prev_c = c
        raise QueryParseError("Unbalanced brackets")

    def parse(self, text):
        """
        Builds a query from the table given and english text expression
        """
        if self.debug:
            print("PARSING", repr(text))
        # build names of possible searchable fields
        fields = {field.name.lower(): field for field in self.table if field.readable}
        # in the stack we put queries and logical operators only
        stack = []

        # match a token using the regex and return the token and left over text
        def next(text, regex, ignore=False):
            text = text.strip()
            if not text:
                if ignore:
                    return None, text
                raise QueryParseError("Unable to parse truncated expression")
            match = regex.match(text)
            if not match:
                if ignore:
                    return None, text
                raise QueryParseError(f"Unable to parse {text}")
            is_quoted = text[:1] == '"'
            token, text = match.group(1), match.group(2)
            # if text is not quoted, remove duplicated spaces
            if is_quoted:
                token = token[1:-1]
            else:
                token = self.re_spaces.sub(" ", token)
            if self.debug:
                print("MATCH", repr(token), repr(text))
            return token, text.strip()

        # loop until there is more text to process
        while text.strip():
            negate = False
            # match if we start with "not"
            token, text = next(text, self.re_not, ignore=True)
            if token:
                negate = True
            # deal with a nested expression or parse the current expression
            if text.startswith("("):
                i = QueryBuilder._find_closing_bracket(text)
                token, text = text[1:i], text[i + 1 :].strip()
                query = self.parse(token)
            else:
                token, text = next(text, self.re_token)
                # match a field name
                if token.lower() not in self.fields:
                    raise QueryParseError(
                        f"Unable to parse {token}, expected a field name"
                    )
                field = self.fields[token]
                is_text = field.type in ("string", "text", "blob")
                has_contains = is_text or field.type.startswith("list:")
                # match an operator
                token, text = next(text, self.re_op)
                token = self.tokens_ops[token]
                # if the operator is a field modifier, get the next operator
                if is_text and token == "lower":
                    token, text = next(text, self.re_op)
                    token = self.tokens_ops.get(token)
                    field = field.lower()
                elif is_text and token == "upper":
                    token, text = next(text, self.re_op)
                    token = self.tokens_ops.get(token)
                    field = field.upper()
                if token == "is null":
                    query = field == None
                elif token == "is not null":
                    query = field != None
                elif field.type == "boolean" and token == "is true":
                    query = field == True
                elif field.type == "boolean" and token == "is false":
                    query = field == False
                else:
                    # the operator requires a value, match a value
                    value, text = next(text, self.re_value)
                    validate(field, value)
                    if token == "==":
                        query = field == value
                    elif token == "!=":
                        query = field != value
                    elif token == "<":
                        query = field < value
                    elif token == ">":
                        query = field > value
                    elif token == "<=":
                        query = field <= value
                    elif token == ">=":
                        query = field >= value
                    elif has_contains and token == "contains":
                        query = field.contains(value)
                    elif is_text and token == "startswith":
                        query = field.startswith(value)
                    elif token == "belongs" and field:
                        # the operator allows multiple values, macth next values
                        value = [value]
                        while text.startswith(","):
                            token, text = next(text[1:], self.re_value)
                            validate(field, token)
                            value.append(token)
                        query = field.belongs(value)
                    else:
                        raise QueryParseError(
                            f"Unable to parse {token}, expected an operator"
                        )
            # we mad matched a not so negate the whole expression
            if negate:
                query = ~query
            # we have a query, put it in stack or combine it into a logical expression
            if len(stack) > 1 and stack[-1] == "and":
                stack[-2:] = [stack[-2] & query]
            elif len(stack) > 1 and stack[-1] == "or":
                stack[-2:] = [stack[-2] | query]
            else:
                stack.append(query)

            # we have a query match next "and" or "or" and put them in stack
            token, text = next(text, self.re_token, ignore=True)
            if not token:
                break
            elif token in self.tokens_and_or and text:
                stack.append(self.tokens_and_or[token])
            else:
                raise QueryParseError(f"Unable to parse {token}, expected and/or")
        if len(stack) > 1:
            # this should never happen
            raise QueryParseError("Internal error: leftover stack")
        # return the one query left in stack
        return stack[-1] if stack else self.table.id > 0



import unittest

from pydal import DAL, Field
from pydal.tools.tags import Tags


class TestTags(unittest.TestCase):
    def test_tags(self):
        db = DAL("sqlite:memory")
        db.define_table("thing", Field("name"))
        properties = Tags(db.thing)
        id1 = db.thing.insert(name="chair")
        id2 = db.thing.insert(name="table")
        properties.add(id1, "color/red")
        properties.add(id1, "style/modern")
        properties.add(id2, "color/green")
        properties.add(id2, "material/wood")

        self.assertTrue(properties.get(id1), ["color/red", "style/modern"])
        self.assertTrue(properties.get(id2), ["color/green", "material/wood"])

        rows = db(properties.find(["style/modern"])).select()
        self.assertTrue(rows.first().id, id1)

        rows = db(properties.find(["material/wood"])).select()
        self.assertTrue(rows.first().id, id1)

        rows = db(properties.find(["color"])).select()
        self.assertTrue(len(rows), 2)



import threading

GLOBAL_LOCKER = threading.RLock()
THREAD_LOCAL = threading.local()

DEFAULT = lambda: None


def IDENTITY(x):
    return x


def OR(a, b):
    return a | b


def AND(a, b):
    return a & b



import re

from .._compat import PY2, integer_types, iteritems, long, to_unicode
from .._globals import IDENTITY
from ..utils import split_uri_args
from . import adapters, with_connection_or_raise
from .base import SQLAdapter


class Slicer(object):
    def rowslice(self, rows, minimum=0, maximum=None):
        if maximum is None:
            return rows[minimum:]
        return rows[minimum:maximum]


class MSSQL(SQLAdapter):
    dbengine = "mssql"
    drivers = ("pyodbc", "pytds", "pymssql", "mssql-python")

    REGEX_DSN = "^.+$"
    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]+|\[[^\]]+\])(:(?P<port>\d+))?"
        "/(?P<db>[^?]+)"
        r"(\?(?P<uriargs>.*))?$"
    )

    def __init__(
        self,
        db,
        uri,
        pool_size=0,
        folder=None,
        db_codec="UTF-8",
        credential_decoder=IDENTITY,
        driver_args={},
        adapter_args={},
        srid=4326,
        after_connection=None,
    ):
        self.srid = srid
        super(MSSQL, self).__init__(
            db,
            uri,
            pool_size,
            folder,
            db_codec,
            credential_decoder,
            driver_args,
            adapter_args,
            after_connection,
        )

    def _initialize_(self):
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        if "@" not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            self.dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)
            user = self.credential_decoder(m.group("user"))
            password = self.credential_decoder(m.group("password"))
            if password is None:
                password = ""
            host = m.group("host")
            db = m.group("db")
            port = m.group("port") or "1433"
            # Parse the optional uri name-value arg pairs after the '?'
            # (in the form of arg1=value1&arg2=value2&...)
            # (drivers like FreeTDS insist on uppercase parameter keys)
            argsdict = {"DRIVER": "{SQL Server}"}
            uriargs = m.group("uriargs")
            if uriargs:
                for argkey, argvalue in split_uri_args(
                    uriargs, separators="&", need_equal=True
                ).items():
                    argsdict[argkey.upper()] = argvalue
            uriargs = ";".join(["%s=%s" % (ak, av) for (ak, av) in iteritems(argsdict)])
            self.dsn = "SERVER=%s;PORT=%s;DATABASE=%s;UID=%s;PWD=%s;%s" % (
                host,
                port,
                db,
                user,
                password,
                uriargs,
            )

    def connector(self):
        return self.driver.connect(self.dsn, **self.driver_args)

    def lastrowid(self, table):
        self.execute("SELECT SCOPE_IDENTITY();")
        return int(self.cursor.fetchone()[0])


@adapters.register_for("mssql")
class MSSQL1(MSSQL, Slicer):
    pass


@adapters.register_for("mssql3")
class MSSQL3(MSSQL):
    pass


@adapters.register_for("mssql4")
class MSSQL4(MSSQL):
    pass


class MSSQLN(MSSQL):
    def represent(self, obj, field_type):
        rv = super(MSSQLN, self).represent(obj, field_type)
        if field_type in ("string", "text", "json") and rv.startswith("'"):
            rv = "N" + rv
        return rv

    @with_connection_or_raise
    def execute(self, *args, **kwargs):
        if PY2:
            args = list(args)
            args[0] = to_unicode(args[0])
        return super(MSSQLN, self).execute(*args, **kwargs)


@adapters.register_for("mssqln", "mssql2")
class MSSQL1N(MSSQLN, Slicer):
    pass


@adapters.register_for("mssql3n")
class MSSQL3N(MSSQLN):
    pass


@adapters.register_for("mssql4n")
class MSSQL4N(MSSQLN):
    pass


@adapters.register_for("pytds")
class PyTDS(MSSQL):
    def _initialize_(self):
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        if "@" not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            self.dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)
            self.dsn = m.group("host")
            self.driver_args.update(
                user=self.credential_decoder(m.group("user")),
                password=self.credential_decoder(m.group("password")) or "",
                database=m.group("db"),
                port=m.group("port") or "1433",
            )

    def connector(self):
        return self.driver.connect(self.dsn, **self.driver_args)


@adapters.register_for("vertica")
class Vertica(MSSQL1):
    def lastrowid(self, table):
        self.execute("SELECT SCOPE_IDENTITY();")
        return int(self.cursor.fetchone()[0])


@adapters.register_for("sybase")
class Sybase(MSSQL1):
    dbengine = "sybase"

    def _initialize_(self):
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        if "@" not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)
            user = self.credential_decoder(m.group("user"))
            password = self.credential_decoder(m.group("password"))
            if password is None:
                password = ""
            host = m.group("host")
            db = m.group("db")
            port = m.group("port") or "1433"
            self.dsn = "sybase:host=%s:%s;dbname=%s" % (host, port, db)
            self.driver_args.update(
                user=self.credential_decoder(user),
                passwd=self.credential_decoder(password),
            )


# Added support for Pymssql
@adapters.register_for("pymssql")
class PyMssql(MSSQL):
    def _initialize_(self):
        import pymssql

        self.driver = pymssql
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        if "@" not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            self.dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)
            self.dsn = m.group("host")
            self.driver_args.update(
                user=self.credential_decoder(m.group("user")),
                password=self.credential_decoder(m.group("password")) or "",
                database=m.group("db"),
                port=m.group("port") or "1433",
            )

    def connector(self):
        return self.driver.connect(self.dsn, **self.driver_args)


# Microsoft mssql-python driver support (SQL Server 2012+)
@adapters.register_for("mssqlpython")
class MSSQLPython(MSSQL4):
    """
    MSSQL adapter for Microsoft's official mssql-python driver.

    Supports SQL Server 2012+ with modern OFFSET/FETCH pagination.
    Features encryption-by-default and Microsoft Entra ID authentication.

    URI format:
        mssqlpython://user:password@server:port/database?param=value

    Supported URI parameters:
        - encrypt: Enable encryption (default: yes, values: yes/no/true/false/1/0)
        - trust_server_certificate: Trust self-signed certificates (default: no)
        - authentication: Authentication mode. Supported values:
            * ActiveDirectoryPassword: Username/password authentication
            * ActiveDirectoryInteractive: Interactive browser login
            * ActiveDirectoryMSI: Managed Identity (system/user-assigned)
            * ActiveDirectoryServicePrincipal: Service principal with client ID/secret
            * ActiveDirectoryIntegrated: Integrated Windows Auth/Kerberos
            * ActiveDirectoryDefault: Default authentication based on environment
            * ActiveDirectoryDeviceCode: Device code flow authentication

    Example URIs:
        # Standard SQL authentication with encryption
        mssqlpython://user:pass@server.database.windows.net:1433/mydb

        # Entra ID Interactive authentication
        mssqlpython://user@server.database.windows.net/mydb?authentication=ActiveDirectoryInteractive

        # Managed Identity (no user/password needed)
        mssqlpython://@server.database.windows.net/mydb?authentication=ActiveDirectoryMSI

        # Disable encryption (for testing/legacy)
        mssqlpython://user:pass@localhost:1433/mydb?encrypt=no&trust_server_certificate=yes
    """

    drivers = (
        "mssql-python",
    )  # Note: URI scheme is mssqlpython://, driver name is mssql-python

    # Override REGEX_URI to make user optional for Managed Identity scenarios
    REGEX_URI = (
        r"^(?P<user>[^:@]*)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]+|\[[^\]]+\])(:(?P<port>\d+))?"
        r"/(?P<db>[^?]+)"
        r"(\?(?P<uriargs>.*))?$"
    )

    def _initialize_(self):
        import mssql_python

        self.driver = mssql_python
        # Skip MSSQL._initialize_() and call grandparent to avoid base MSSQL's REGEX
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)

        # Parse URI query parameters
        uriargs = m.group("uriargs")
        uri_args = split_uri_args(uriargs, need_equal=True) if uriargs else {}

        # Extract basic connection parameters
        user = self.credential_decoder(m.group("user")) if m.group("user") else ""
        password = (
            self.credential_decoder(m.group("password")) if m.group("password") else ""
        )

        # Build driver connection arguments
        self.driver_args.update(
            server=m.group("host"),
            database=m.group("db"),
            port=int(m.group("port") or "1433"),
        )

        # Add credentials if provided (not required for some Entra ID auth modes)
        if user:
            self.driver_args["user"] = user
        if password:
            self.driver_args["password"] = password

        # Handle authentication parameter
        # Supported modes: ActiveDirectoryPassword, ActiveDirectoryInteractive,
        # ActiveDirectoryMSI, ActiveDirectoryServicePrincipal,
        # ActiveDirectoryIntegrated, ActiveDirectoryDefault, ActiveDirectoryDeviceCode
        if "authentication" in uri_args:
            self.driver_args["authentication"] = uri_args["authentication"]

        # Handle encryption parameters (default: enabled for security)
        encrypt_param = uri_args.get("encrypt", "yes").lower()
        # mssql-python expects string values "yes"/"no", not boolean
        if encrypt_param in ("yes", "true", "1"):
            self.driver_args["encrypt"] = "yes"
        else:
            self.driver_args["encrypt"] = "no"

        # Handle trust_server_certificate parameter (default: disabled for security)
        if "trust_server_certificate" in uri_args:
            trust_param = uri_args["trust_server_certificate"].lower()
            # mssql-python expects string values "yes"/"no", not boolean
            if trust_param in ("yes", "true", "1"):
                self.driver_args["trust_server_certificate"] = "yes"
            else:
                self.driver_args["trust_server_certificate"] = "no"

        # Pass through any additional driver-specific parameters
        # Exclude already-handled parameters
        handled_params = {"authentication", "encrypt", "trust_server_certificate"}
        for key, value in uri_args.items():
            if key not in handled_params:
                self.driver_args[key] = value

    def connector(self):
        return self.driver.connect(**self.driver_args)



import collections
import copy
import fnmatch
import functools
import re
import traceback

from .utils import utcnow

__version__ = "0.1"

__all__ = ["RestAPI", "Policy", "ALLOW_ALL_POLICY", "DENY_ALL_POLICY"]

MAX_LIMIT = 1000


class PolicyViolation(ValueError):
    pass


class InvalidFormat(ValueError):
    pass


class NotFound(ValueError):
    pass


def maybe_call(value):
    "call value if callable else return value"
    return value() if callable(value) else value


def trydo(f, default):
    "return f() if no exception else return default"
    try:
        return f()
    except Exception:
        return default


def error_wrapper(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        data = {}
        try:
            data = func(*args, **kwargs)
            if not data.get("errors"):
                data["status"] = "success"
                data["code"] = 200
            else:
                data["status"] = "error"
                data["message"] = "Validation Errors"
                data["code"] = 422
        except PolicyViolation as e:
            data["status"] = "error"
            data["message"] = str(e)
            data["code"] = 401
        except NotFound as e:
            data["status"] = "error"
            data["message"] = str(e)
            data["code"] = 404
        except (InvalidFormat, KeyError, ValueError) as e:
            data["status"] = "error"
            data["message"] = str(e)
            data["code"] = 400
        finally:
            data["timestamp"] = utcnow().isoformat()
            data["api_version"] = __version__
        return data

    return wrapper


class Policy(object):
    model = {
        "POST": {"authorize": False, "fields": None},
        "PUT": {"authorize": False, "fields": None},
        "DELETE": {"authorize": False},
        "GET": {
            "authorize": False,
            "fields": None,
            "query": None,
            "allowed_patterns": [],
            "denied_patterns": [],
            "limit": MAX_LIMIT,
            "allow_lookup": False,
        },
    }

    def __init__(self):
        self.info = {}

    def set(self, tablename, method="GET", **attributes):
        method = method.upper()
        if not method in self.model:
            raise InvalidFormat("Invalid policy method: %s" % method)
        invalid_keys = [key for key in attributes if key not in self.model[method]]
        if invalid_keys:
            raise InvalidFormat("Invalid keys: %s" % ",".join(invalid_keys))
        if not tablename in self.info:
            self.info[tablename] = copy.deepcopy(self.model)
        self.info[tablename][method].update(attributes)

    def get(self, tablename, method, name):
        policy = self.info.get(tablename) or self.info.get("*")
        if not policy:
            raise PolicyViolation("No policy for this object")
        return maybe_call(policy[method][name])

    def check_if_allowed(
        self, method, tablename, id=None, get_vars=None, post_vars=None, exceptions=True
    ):
        get_vars = get_vars or {}
        post_vars = post_vars or {}
        policy = self.info.get(tablename) or self.info.get("*")
        if not policy:
            if exceptions:
                raise PolicyViolation("No policy for this object")
            return False
        policy = policy.get(method.upper())
        if not policy:
            if exceptions:
                raise PolicyViolation("No policy for this method")
            return False
        authorize = policy.get("authorize")
        if authorize is False or (
            callable(authorize) and not authorize(tablename, id, get_vars, post_vars)
        ):
            if exceptions:
                raise PolicyViolation("Not authorized")
            return False
        for key in get_vars:
            if any(fnmatch.fnmatch(key, p) for p in policy["denied_patterns"]):
                if exceptions:
                    raise PolicyViolation("Pattern is not allowed")
                return False
            allowed_patterns = policy["allowed_patterns"]
            if "**" not in allowed_patterns and not any(
                fnmatch.fnmatch(key, p) for p in allowed_patterns
            ):
                if exceptions:
                    raise PolicyViolation("Pattern is not explicitely allowed")
                return False
        return True

    def check_if_lookup_allowed(self, tablename, exceptions=True):
        policy = self.info.get(tablename) or self.info.get("*")
        if not policy:
            if exceptions:
                raise PolicyViolation("No policy for this object")
            return False
        policy = policy.get("GET")
        if not policy:
            if exceptions:
                raise PolicyViolation("No policy for this method")
            return False
        if policy.get("allow_lookup"):
            return True
        return False

    def allowed_fieldnames(self, table, method="GET"):
        method = method.upper()
        policy = self.info.get(table._tablename) or self.info.get("*", {})
        policy = policy[method]
        allowed_fieldnames = policy.get("fields")
        if allowed_fieldnames is None:
            allowed_fieldnames = [
                f.name
                for f in table
                if (method == "GET" and maybe_call(f.readable))
                or (method != "GET" and maybe_call(f.writable))
            ]
        return allowed_fieldnames

    def check_fieldnames(self, table, fieldnames, method="GET"):
        allowed_fieldnames = self.allowed_fieldnames(table, method)
        invalid_fieldnames = set(fieldnames) - set(allowed_fieldnames)
        if invalid_fieldnames:
            raise InvalidFormat("Invalid fields: %s" % list(invalid_fieldnames))


DENY_ALL_POLICY = Policy()
ALLOW_ALL_POLICY = Policy()
ALLOW_ALL_POLICY.set(
    tablename="*",
    method="GET",
    authorize=True,
    allowed_patterns=["**"],
    allow_lookup=True,
)
ALLOW_ALL_POLICY.set(tablename="*", method="POST", authorize=True)
ALLOW_ALL_POLICY.set(tablename="*", method="PUT", authorize=True)
ALLOW_ALL_POLICY.set(tablename="*", method="DELETE", authorize=True)


class RestAPI(object):
    re_table_and_fields = re.compile(r"\w+([\w+(,\w+)+])?")
    re_lookups = re.compile(
        r"((\w*\!?\:)?(\w+(\[\w+(,\w+)*\])?)(\.\w+(\[\w+(,\w+)*\])?)*)"
    )
    re_no_brackets = re.compile(r"\[.*?\]")

    def __init__(self, db, policy):
        self.db = db
        self.policy = policy
        self.allow_count = "legacy"

    @error_wrapper
    def __call__(
        self,
        method,
        tablename,
        id=None,
        get_vars=None,
        post_vars=None,
        allow_count="legacy",
    ):
        method = method.upper()
        get_vars = get_vars or {}
        post_vars = post_vars or {}
        self.allow_count = allow_count
        # validate incoming request
        tname, tfieldnames = RestAPI.parse_table_and_fields(tablename)
        if not tname in self.db.tables:
            raise InvalidFormat("Invalid table name: %s" % tname)
        if self.policy:
            self.policy.check_if_allowed(method, tablename, id, get_vars, post_vars)
            if method in ["POST", "PUT"]:
                self.policy.check_fieldnames(
                    self.db[tablename], post_vars.keys(), method
                )
        # apply rules
        if method == "GET":
            if id:
                get_vars["id.eq"] = id
            return self.search(tablename, get_vars)
        elif method == "POST":
            table = self.db[tablename]
            return table.validate_and_insert(**post_vars)

        elif method == "PUT":
            id = id or post_vars["id"]
            if not id:
                raise InvalidFormat("No item id specified")
            table = self.db[tablename]
            data = table.validate_and_update(id, **post_vars)
            if not data.get("errors") and not data.get("updated"):
                raise NotFound("Item not found")
            return data
        elif method == "DELETE":
            id = id or post_vars["id"]
            if not id:
                raise InvalidFormat("No item id specified")
            table = self.db[tablename]
            deleted = self.db(table._id == id).delete()
            if not deleted:
                raise NotFound("Item not found")
            return {"deleted": deleted}

    def table_model(self, table, fieldnames):
        """converts a table into its form template"""
        items = []
        fields = post_fields = put_fields = table.fields
        if self.policy:
            fields = self.policy.allowed_fieldnames(table, method="GET")
            put_fields = self.policy.allowed_fieldnames(table, method="PUT")
            post_fields = self.policy.allowed_fieldnames(table, method="POST")
        for fieldname in fields:
            if fieldnames and not fieldname in fieldnames:
                continue
            field = table[fieldname]
            item = {"name": field.name, "label": field.label}
            # https://github.com/collection-json/extensions/blob/master/template-validation.md
            item["default"] = (
                field.default() if callable(field.default) else field.default
            )
            parts = field.type.split()
            item["type"] = parts[0].split("(")[0]
            if len(parts) > 1:
                item["references"] = parts[1]
            if hasattr(field, "regex"):
                item["regex"] = field.regex
            item["required"] = field.required
            item["unique"] = field.unique
            item["post_writable"] = field.name in post_fields
            item["put_writable"] = field.name in put_fields
            item["options"] = field.options
            if field.type == "id":
                item["referenced_by"] = [
                    "%s.%s" % (f._tablename, f.name)
                    for f in table._referenced_by
                    if self.policy
                    and self.policy.check_if_allowed(
                        "GET", f._tablename, exceptions=False
                    )
                ]
            items.append(item)
        return items

    @staticmethod
    def make_query(field, condition, value):
        expression = {
            "eq": lambda: field == value,
            "ne": lambda: field != value,
            "lt": lambda: field < value,
            "gt": lambda: field > value,
            "le": lambda: field <= value,
            "ge": lambda: field >= value,
            "startswith": lambda: field.startswith(str(value)),
            "in": lambda: field.belongs(
                value.split(",") if isinstance(value, str) else list(value)
            ),
            "contains": lambda: field.contains(value),
        }
        return expression[condition]()

    @staticmethod
    def parse_table_and_fields(text):
        if not RestAPI.re_table_and_fields.match(text):
            raise ValueError
        parts = text.split("[")
        if len(parts) == 1:
            return parts[0], []
        elif len(parts) == 2:
            return parts[0], parts[1][:-1].split(",")

    def search(self, tname, vars):
        def check_table_permission(tablename):
            if self.policy:
                self.policy.check_if_allowed("GET", tablename)

        def check_table_lookup_permission(tablename):
            if self.policy:
                self.policy.check_if_lookup_allowed(tablename)

        def filter_fieldnames(table, fieldnames):
            if self.policy:
                if fieldnames:
                    self.policy.check_fieldnames(table, fieldnames)
                else:
                    fieldnames = self.policy.allowed_fieldnames(table)
            elif not fieldnames:
                fieldnames = table.fields
            return fieldnames

        db = self.db
        tname, tfieldnames = RestAPI.parse_table_and_fields(tname)
        check_table_permission(tname)
        tfieldnames = filter_fieldnames(db[tname], tfieldnames)
        query = []
        offset = 0
        limit = 100
        model = False
        options_list = False
        table = db[tname]
        queries = []
        if self.policy:
            common_query = self.policy.get(tname, "GET", "query")
            if common_query:
                queries.append(common_query)
        hop1 = collections.defaultdict(list)
        hop2 = collections.defaultdict(list)
        hop3 = collections.defaultdict(list)
        model_fieldnames = tfieldnames
        lookup = {}
        orderby = None
        do_count = False
        for key, value in vars.items():
            if key == "@offset":
                offset = int(value)
            elif key == "@limit":
                limit = min(
                    int(value),
                    (
                        self.policy.get(tname, "GET", "limit")
                        if self.policy
                        else MAX_LIMIT
                    ),
                )
            elif key == "@order":
                orderby = [
                    ~table[f[1:]] if f[:1] == "~" else table[f]
                    for f in value.split(",")
                    if f.lstrip("~") in table.fields
                ] or None
            elif key == "@lookup":
                lookup = {item[0]: {} for item in RestAPI.re_lookups.findall(value)}
            elif key == "@model":
                model = str(value).lower()[:1] == "t"
            elif key == "@options_list":
                options_list = str(value).lower()[:1] == "t"
            elif key == "@count":
                if self.allow_count:
                    do_count = str(value).lower()[:1] == "t"
            else:
                key_parts = key.rsplit(".")
                if not key_parts[-1] in (
                    "eq",
                    "ne",
                    "gt",
                    "lt",
                    "ge",
                    "le",
                    "startswith",
                    "contains",
                    "in",
                ):
                    key_parts.append("eq")
                is_negated = key_parts[0] == "not"
                if is_negated:
                    key_parts = key_parts[1:]
                key, condition = key_parts[:-1], key_parts[-1]
                if len(key) == 1:  # example: name.eq=='Chair'
                    query = self.make_query(table[key[0]], condition, value)
                    queries.append(query if not is_negated else ~query)
                elif len(key) == 2:  # example: color.name.eq=='red'
                    hop1[is_negated, key[0]].append((key[1], condition, value))
                elif len(key) == 3:  # example: a.rel.desc.eq=='above'
                    hop2[is_negated, key[0], key[1]].append((key[2], condition, value))
                elif len(key) == 4:  # example: a.rel.b.name.eq == 'Table'
                    hop3[is_negated, key[0], key[1], key[2]].append(
                        (key[3], condition, value)
                    )

        for item in hop1:
            is_negated, fieldname = item
            ref_tablename = table[fieldname].type.split(" ")[1]
            ref_table = db[ref_tablename]
            subqueries = [self.make_query(ref_table[k], c, v) for k, c, v in hop1[item]]
            subquery = functools.reduce(lambda a, b: a & b, subqueries)
            query = table[fieldname].belongs(db(subquery)._select(ref_table._id))
            queries.append(query if not is_negated else ~query)

        for item in hop2:
            is_negated, linkfield, linktable = item
            ref_table = db[linktable]
            subqueries = [self.make_query(ref_table[k], c, v) for k, c, v in hop2[item]]
            subquery = functools.reduce(lambda a, b: a & b, subqueries)
            query = table._id.belongs(db(subquery)._select(ref_table[linkfield]))
            queries.append(query if not is_negated else ~query)

        for item in hop3:
            is_negated, linkfield, linktable, otherfield = item
            ref_table = db[linktable]
            ref_ref_tablename = ref_table[otherfield].type.split(" ")[1]
            ref_ref_table = db[ref_ref_tablename]
            subqueries = [
                self.make_query(ref_ref_table[k], c, v) for k, c, v in hop3[item]
            ]
            subquery = functools.reduce(lambda a, b: a & b, subqueries)
            subquery &= ref_ref_table._id == ref_table[otherfield]
            query = table._id.belongs(
                db(subquery)._select(ref_table[linkfield], groupby=ref_table[linkfield])
            )
            queries.append(query if not is_negated else ~query)

        if not queries:
            queries.append(table)

        query = functools.reduce(lambda a, b: a & b, queries)
        tfields = [
            table[tfieldname]
            for tfieldname in tfieldnames
            if table[tfieldname].type != "password"
        ]
        passwords = [
            tfieldname
            for tfieldname in tfieldnames
            if table[tfieldname].type == "password"
        ]
        rows = db(query).select(
            *tfields, limitby=(offset, limit + offset), orderby=orderby
        )
        if passwords:
            dpass = {password: "******" for password in passwords}
            for row in rows:
                row.update(dpass)

        lookup_map = {}
        for key in list(lookup.keys()):
            name, key = key.split(":") if ":" in key else ("", key)
            clean_key = RestAPI.re_no_brackets.sub("", key)
            lookup_map[clean_key] = {
                "name": name.rstrip("!") or clean_key,
                "collapsed": name.endswith("!"),
            }
            key = key.split(".")

            if len(key) == 1:
                key, tfieldnames = RestAPI.parse_table_and_fields(key[0])
                ref_tablename = table[key].type.split(" ")[1]
                ref_table = db[ref_tablename]
                tfieldnames = filter_fieldnames(ref_table, tfieldnames)
                check_table_lookup_permission(ref_tablename)
                ids = [row[key] for row in rows]
                tfields = [
                    ref_table[tfieldname]
                    for tfieldname in tfieldnames
                    if ref_table[tfieldname].type != "password"
                ]
                if not "id" in tfieldnames:
                    tfields.append(ref_table["id"])
                drows = db(ref_table._id.belongs(ids)).select(*tfields).as_dict()
                if tfieldnames and not "id" in tfieldnames:
                    for row in drows.values():
                        del row["id"]
                lkey, collapsed = lookup_map[key]["name"], lookup_map[key]["collapsed"]
                for row in rows:
                    new_row = drows.get(row[key])
                    if collapsed:
                        del row[key]
                        for rkey in tfieldnames:
                            row[lkey + "." + rkey] = new_row[rkey] if new_row else None
                    else:
                        row[lkey] = new_row

            elif len(key) == 2:
                lfield, key = key
                key, tfieldnames = RestAPI.parse_table_and_fields(key)
                check_table_lookup_permission(key)
                ref_table = db[key]
                tfieldnames = filter_fieldnames(ref_table, tfieldnames)
                ids = [row["id"] for row in rows]
                tfields = [ref_table[tfieldname] for tfieldname in tfieldnames]
                if not lfield in tfieldnames:
                    tfields.append(ref_table[lfield])
                lrows = db(ref_table[lfield].belongs(ids)).select(*tfields)
                drows = collections.defaultdict(list)
                for row in lrows:
                    row = row.as_dict()
                    drows[row[lfield]].append(row)
                    if not lfield in tfieldnames:
                        del row[lfield]
                lkey = lookup_map[lfield + "." + key]["name"]
                for row in rows:
                    row[lkey] = drows.get(row.id, [])

            elif len(key) == 3:
                lfield, key, rfield = key
                key, tfieldnames = RestAPI.parse_table_and_fields(key)
                rfield, tfieldnames2 = RestAPI.parse_table_and_fields(rfield)
                check_table_lookup_permission(key)
                ref_table = db[key]
                ref_ref_tablename = ref_table[rfield].type.split(" ")[1]
                check_table_lookup_permission(ref_ref_tablename)
                ref_ref_table = db[ref_ref_tablename]
                tfieldnames = filter_fieldnames(ref_table, tfieldnames)
                tfieldnames2 = filter_fieldnames(ref_ref_table, tfieldnames2)
                ids = [row["id"] for row in rows]
                tfields = [ref_table[tfieldname] for tfieldname in tfieldnames]
                if not lfield in tfieldnames:
                    tfields.append(ref_table[lfield])
                if not rfield in tfieldnames:
                    tfields.append(ref_table[rfield])
                tfields += [ref_ref_table[tfieldname] for tfieldname in tfieldnames2]
                left = ref_ref_table.on(ref_table[rfield] == ref_ref_table["id"])
                lrows = db(ref_table[lfield].belongs(ids)).select(*tfields, left=left)
                drows = collections.defaultdict(list)
                lkey = lfield + "." + key + "." + rfield
                lkey, collapsed = (
                    lookup_map[lkey]["name"],
                    lookup_map[lkey]["collapsed"],
                )
                for row in lrows:
                    row = row.as_dict()
                    new_row = row[key]
                    lfield_value, rfield_value = new_row[lfield], new_row[rfield]
                    if not lfield in tfieldnames:
                        del new_row[lfield]
                    if not rfield in tfieldnames:
                        del new_row[rfield]
                    if collapsed:
                        new_row.update(row[ref_ref_tablename])
                    else:
                        new_row[rfield] = row[ref_ref_tablename]
                    drows[lfield_value].append(new_row)
                for row in rows:
                    row[lkey] = drows.get(row.id, [])

        response = {}
        if not options_list:
            response["items"] = rows.as_list()
        else:
            if callable(table._format):
                f = lambda row: trydo(lambda: table._format(row), str(row.id))
            elif table._format:
                f = lambda row: trydo(lambda: table._format % row, str(row.id))
            else:
                f = lambda row: str(row.id)
            response["items"] = [dict(value=row.id, text=f(row)) for row in rows]

        if do_count or (self.allow_count == "legacy" and offset == 0):
            response["count"] = db(query).count()
        if model:
            response["model"] = self.table_model(table, model_fieldnames)
        return response
